# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Iterator, List

from sqlalchemy import ForeignKey, Text, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import (
    Mapped,
    declared_attr,
    mapped_column,
    relationship
)

from ..model import Model


@dataclass(frozen=True, kw_only=True)
class PipelineStepModels(IterableABC[type[Model]]):
    """
    Contains the needed models for users to run steps
    in validation pipelines.

    Additionally contains `_user_mixin`, from which developers
    should inherit their `User` class from `tol.sql.auth`.
    """

    pipeline: type[Model]
    pipeline_step: type[Model]
    upload: type[Model]

    _user_mixin: type[Any]

    def __iter__(self) -> Iterator[type[Model]]:
        """
        Returns in order they should be deleted
        """
        return iter(
            [
                self.upload,
                self.pipeline_step,
                self.pipeline,
            ]
        )


def create_pipeline_step_models(
    base_model_class: type[Model]
) -> PipelineStepModels:
    """
    Creates all needed models (and joining tables) for
    validation pipelines.

    Returns a `PipelineStepModels` instance that functions like an
    `Iterable`.
    """

    class Pipeline(base_model_class):
        __tablename__ = 'pipeline'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        name: Mapped[str] = mapped_column(
            nullable=False,
            unique=True,
        )

        config: Mapped[dict[str, Any]] = mapped_column(
            nullable=False,
            default={}
        )

        uploads: Mapped[list[Upload]] = relationship(
            'Upload',
            back_populates='pipeline'
        )

        steps: Mapped[list[PipelineStep]] = relationship(
            'PipelineStep',
            back_populates='pipeline'
        )

    class PipelineStep(base_model_class):
        __tablename__ = 'pipeline_steps'
        __table_args__ = (
            UniqueConstraint(
                'pipeline_id',
                'stage',
                'step_order',
                name='uq_pipeline_step_stage_order'
            ),
        )

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        pipeline_id: Mapped[int] = mapped_column(
            ForeignKey('pipeline.id')
        )

        step_name: Mapped[str] = mapped_column(
            nullable=False
        )

        stage: Mapped[int] = mapped_column(
            nullable=False
        )

        step_order: Mapped[int] = mapped_column(
            nullable=False
        )

        description: Mapped[str | None] = mapped_column(
            Text,
            nullable=True
        )

        is_visible: Mapped[bool] = mapped_column(
            nullable=False,
            default=True
        )

        config: Mapped[dict[str, Any]] = mapped_column(
            JSONB,
            nullable=False,
            default={}
        )

        pipeline: Mapped[Pipeline] = relationship(
            'Pipeline',
            back_populates='steps'
        )

    class Upload(base_model_class):
        __tablename__ = 'upload'

        id: Mapped[int] = mapped_column(  # noqa A003
            primary_key=True,
            autoincrement=True
        )

        s3_bucket: Mapped[str] = mapped_column(nullable=False,)
        s3_filename: Mapped[str] = mapped_column(nullable=False)
        spreadsheet_config: Mapped[str] = mapped_column(nullable=True)

        user_id: Mapped[int] = mapped_column(
            ForeignKey('user.id'),
            nullable=True
        )

        pipeline_id: Mapped[int] = mapped_column(
            ForeignKey('pipeline.id'),
            nullable=False
        )

        destination: Mapped[str] = mapped_column(nullable=False)
        flow_run_id: Mapped[str] = mapped_column(nullable=True)

        date_started: Mapped[datetime] = mapped_column(
            nullable=False,
            default=datetime.now
        )

        validation_results: Mapped[List[Dict[str, Any]]] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        completed: Mapped[bool] = mapped_column(
            nullable=False,
            default=False
        )

        validation_status: Mapped[str] = mapped_column(
            nullable=True,
            default='in_progress'
        )

        rejection_reason: Mapped[str] = mapped_column(
            nullable=True
        )

        hidden: Mapped[str] = mapped_column(
            nullable=True,
            default=False
        )

        upload_name: Mapped[str] = mapped_column(
            nullable=True,
        )

        failure_message: Mapped[str | None] = mapped_column(
            nullable=True,
        )

        pipeline: Mapped['Pipeline'] = relationship(
            back_populates='uploads',
            foreign_keys=[pipeline_id]
        )

        user: Mapped['User'] = relationship(  # noqa F821
            back_populates='user_uploads',
            foreign_keys=[user_id]
        )

    class _UserMixin:

        @declared_attr
        def user_uploads(self) -> Mapped[list[Upload]]:
            return relationship(
                'Upload',
                back_populates='user'
            )

    return PipelineStepModels(
        pipeline=Pipeline,
        pipeline_step=PipelineStep,
        upload=Upload,
        _user_mixin=_UserMixin
    )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import (
    Iterable
)

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticSampleToBoldSampleConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        loc = f'{data_object.sts_gal_name} - {data_object.sts_gal_abbreviation}'
        if len(loc) == 51:
            loc = f'{data_object.sts_gal_name}- {data_object.sts_gal_abbreviation}'
        if len(loc) > 50:
            loc = 'dummy'
        if data_object.sts_gal_name is None:
            loc = ''
        if data_object.sts_rackid is not None:
            project_code = data_object.sts_rackid.split('_')[0]
        else:
            project_code = ''
        upd_codes = {
            'NHMG': 'BNHMG',
            'CARM': 'BCARM',
            'CAMP': 'SCAMP',
            'EWTA': 'BEWTA',
            'NTDB': 'BNTDB',
            'FMED': 'BFMED'
        }
        if project_code in upd_codes.keys():
            project_code = upd_codes[project_code]
        attributes = {
            'bold_recordset_code_arr': project_code,
            'sampleid':
                data_object.sts_specimen.id
                if data_object.sts_specimen is not None else '',
            'fieldid': '',
            'inst': 'Wellcome Sanger Institute',
            'phylum': 'Arthropoda',
            'class': '',
            'order': '',
            'short_note': loc,
            'notes': '',
            'voucher_type': '',
            'tissue_type':
                ' | '.join(data_object.sts_organism_part)
                if data_object.sts_organism_part is not None else '',
            'collectors': self.__extract_names_from_contributors(data_object.sts_CONTRIBUTORS),
            'collection_date_start':
                data_object.sts_col_date.strftime('%Y-%m-%d')
                if data_object.sts_col_date is not None else '',
            'country/ocean':
                data_object.sts_COUNTRY_OF_COLLECTION.title()
                if data_object.sts_COUNTRY_OF_COLLECTION is not None else '',
            'province/state': '',
            'coord':
                f'{data_object.sts_latitude},{data_object.sts_longitude}'
                if data_object.sts_latitude is not None
                and data_object.sts_longitude is not None else '',
            'elev': '',
            'elev_accuracy': '',
            'collection_date_end': '',
            'sampling_protocol':
                data_object.sts_COLLECTION_METHOD.replace('_', ' ').title()
                if data_object.sts_COLLECTION_METHOD is not None else '',
        }
        ret = self._data_object_factory(
            'sample',
            data_object.id,
            attributes=attributes
        )
        yield ret

    def __extract_names_from_contributors(self, contributors: str) -> str:
        try:
            c_list = contributors.split('|')
        except AttributeError:
            c_list = ''
        names = [c.split(';')[0] for c in c_list]
        return ', '.join(names)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import List

from tol.core import DataObject, Validator

from .interfaces import Condition, ConditionDict, ConditionEvaluator


class AssertOnConditionValidator(Validator, ConditionEvaluator):
    """
    Validates an incoming stream of `DataObject` instances,
    using a condition to check a specific attrbiute. If this
    condition passes, then the assertions will be run, which must
    all pass.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        condition: ConditionDict
        assertions: List[ConditionDict]

    __slots__ = ['__config']
    __config: Config

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()

        self.__config = config

    def _validate_data_object(self, obj: DataObject) -> None:
        # Check condition atribute
        # (only perform the assertions if the condition passes)
        if self._does_condition_pass(Condition.from_dict(self.__config.condition), obj):
            # Perform each assertion
            for assertion in self.__config.assertions:
                self.__perform_assertion(obj, Condition.from_dict(assertion))

    def __perform_assertion(self, obj: DataObject, assertion: Condition) -> None:
        # There's only an error or warning if the assertion condition fails
        condition_passed, found_value = self._evaluate_condition(assertion, obj)
        if not condition_passed:
            if assertion.is_error:
                self.add_error(
                    object_id=obj.id,
                    detail=f'Expected {assertion} (found value {found_value})',
                    field=assertion.field,
                )
            else:
                self.add_warning(
                    object_id=obj.id,
                    detail=f'Expected {assertion} (found value {found_value})',
                    field=assertion.field,
                )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import wraps
from typing import Any, Callable, Union

from ..core import DataObject, DataSourceError

if typing.TYPE_CHECKING:
    from .api_datasource import ApiDataSource


def _get_object_type(
    object_or_type: Union[str, DataObject],
    direct_object: bool
) -> str:
    """
    Gets the `object_type` either:

    - as directly given (if not `direct_object`)
    - from the `.type` (if `direct_object`)
    """

    return (
        object_or_type.type if direct_object
        else object_or_type
    )


def _validate_type_supported(
    ds: ApiDataSource,
    object_type: str
) -> None:

    if object_type not in ds.supported_types:
        detail = (
            f'The type "{object_type}" is unrecognised.'
        )

        raise DataSourceError(
            title='Unknown Type',
            detail=detail,
            status_code=400
        )


def _validate_operation(
    ds: ApiDataSource,
    object_type: str,
    operation_name: str
) -> None:

    supported_operations = ds.supported_operations.get(
        object_type,
        []
    )

    if operation_name not in supported_operations:
        detail = (
            f'The operation {operation_name} is '
            f'unsupported on {object_type}.'
        )
        raise DataSourceError(
            'Unsupported Operation',
            detail,
            400
        )


def validate(
    operation_name: str,
    direct_object: bool = False,
):
    """
    Performs several pre-flight checks:

    - the operation on `ApiDataSource` is supported
    - the `data_object` type is supported

    Set `direct_object` to `True` if the operation
    gives an instance of `DataObject` as its first
    arg.
    """

    def decorator(operation: Callable) -> Callable:

        @wraps(operation)
        def wrapper(
            ds: ApiDataSource,
            data_object_or_type: Union[DataObject, str],
            *args,
            **kwargs
        ) -> Any:

            object_type = _get_object_type(data_object_or_type, direct_object)

            _validate_type_supported(ds, object_type)
            _validate_operation(ds, object_type, operation_name)

            return operation(ds, data_object_or_type, *args, **kwargs)

        return wrapper

    return decorator


def _check_id_is_not_none(source: DataObject) -> None:
    if source.id is None:
        raise DataSourceError(
            title='Unset Source ID',
            detail='The given object has no value set for its ID.',
            status_code=400
        )


def validate_id(operation: Callable) -> Callable:
    """
    Validates that the given instance of `DataObject`
    has a non-`None` ID.
    """

    @wraps(operation)
    def wrapper(
        ds: ApiDataSource,
        source: DataObject,
        *args,
        **kwargs
    ) -> Any:

        _check_id_is_not_none(source)

        return operation(ds, source, *args, **kwargs)

    return wrapper



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import importlib
from typing import Iterator

from dacite import from_dict

from .data_object import DataObject
from .data_source_attribute_metadata import data_source_attribute_metadata
from .datasource import DataSource
from .datasource_error import DataSourceError
from .datasource_filter import DataSourceFilter
from .relationship import RelationshipConfig


class DataSourceUtils:

    @classmethod
    def get_datasource(
        cls,
        datasource_instance_id: str,
        config_datasource: DataSource
    ) -> DataSource:
        datasource_instance = config_datasource.get_one(
            'data_source_instance',
            datasource_instance_id
        )
        if datasource_instance is None:
            raise DataSourceError(
                f'Datasource instance with id {datasource_instance_id} not found'
            )
        return cls.get_datasource_by_datasource_instance(datasource_instance)

    @classmethod
    def get_datasource_by_name(
        cls,
        name: str,
        **kwargs
    ) -> DataSource:
        module = importlib.import_module(f'tol.sources.{name}')
        class_ = getattr(module, name)
        return class_(**kwargs)

    @classmethod
    def get_datasource_by_datasource_instance(
        cls,
        datasource_instance: DataObject,
        **kwargs
    ) -> DataSource:
        datasource_config = datasource_instance.data_source_config
        new_kwargs = dict(datasource_instance.kwargs) if datasource_instance.kwargs else {}
        if kwargs:
            new_kwargs.update(kwargs)
        if datasource_config:
            relationship_config = cls.get_relationship_config_from_data_source_config(
                datasource_config
            )
            amd = data_source_attribute_metadata(
                datasource_config
            )
            runtime_fields = cls.get_runtime_fields_from_data_source_config(
                datasource_config
            )
            new_kwargs.update({
                'relationship_cfg': relationship_config,
                'attribute_metadata': amd,
                'runtime_fields': runtime_fields
            })
        return DataSourceUtils.get_datasource_by_name(
            datasource_instance.builtin_name,
            **new_kwargs
        )

    @classmethod
    def get_relationship_config_from_data_source_config(
        cls,
        datasource_config: DataObject
    ) -> dict:
        relationship_cfg = {}
        for rel in datasource_config.data_source_config_relationships:
            for obj_type in [rel.object_type, rel.foreign_object_type]:
                if obj_type not in relationship_cfg:
                    relationship_cfg[obj_type] = {
                        'to_one': {},
                        'to_many': {},
                        'foreign_keys': {}
                    }
            relationship_cfg[rel.object_type]['to_one'][rel.name] = rel.foreign_object_type
            relationship_cfg[rel.foreign_object_type]['to_many'][rel.foreign_name] = \
                rel.object_type
            relationship_cfg[rel.foreign_object_type]['foreign_keys'][rel.foreign_name] = \
                f'{rel.name}.id'
        return {
            k: from_dict(data_class=RelationshipConfig, data=v)
            for k, v in relationship_cfg.items()
        }

    @classmethod
    def get_runtime_fields_from_data_source_config(
        cls,
        datasource_config: DataObject
    ) -> dict:
        runtime_fields = {}
        f = DataSourceFilter()
        f.and_ = {
            'runtime_definition': {'exists': {}}
        }
        for dsa in datasource_config.data_source_config_attributes:
            if dsa.runtime_definition is None:
                continue
            if dsa.object_type not in runtime_fields:
                runtime_fields[dsa.object_type] = {}
            if 'function' in dsa.runtime_definition:
                from ..elastic.runtime_fields import RuntimeFields  # Break circular import cycle
                method = getattr(RuntimeFields, dsa.runtime_definition['function'])
                runtime_fields[dsa.object_type][dsa.name] = \
                    method(**dsa.runtime_definition.get('function_kwargs', {}))
            if 'script' in dsa.runtime_definition:
                runtime_fields[dsa.object_type][dsa.name] = {
                    'type': dsa.runtime_definition.get('type', 'keyword'),
                    'script': {'source': dsa.runtime_definition['script']}
                }
        return runtime_fields

    @classmethod
    def get_ids(
        cls,
        datasource: DataSource,
        object_type: str,
        id_attribute: str,
        object_filters: DataSourceFilter = None
    ) -> Iterator[str]:
        try:
            yield from cls.__get_ids_via_group_stats(
                datasource,
                object_type,
                id_attribute,
                object_filters
            )
        except (DataSourceError, AttributeError):
            # If the datasource does not support group stats, we will
            # fall back to get_list
            yield from cls.__get_ids_via_get_list(
                datasource,
                object_type,
                id_attribute,
                object_filters
            )

    @classmethod
    def __get_ids_via_group_stats(
        cls,
        datasource: DataSource,
        object_type: str,
        id_attribute: str,
        object_filters: DataSourceFilter = None
    ) -> Iterator[str]:
        uniques = datasource.get_group_stats(
            object_type,
            group_by=[id_attribute],
            stats_fields=[],
            stats=[],
            object_filters=object_filters
        )
        for unique in uniques:
            yield unique['key'][id_attribute]

    @classmethod
    def __get_ids_via_get_list(
        cls,
        datasource: DataSource,
        object_type: str,
        id_attribute: str,
        object_filters: DataSourceFilter = None
    ) -> Iterator[str]:
        ids_seen = set()
        objs = datasource.get_list(
            object_type,
            object_filters=object_filters
        )
        for obj in objs:
            id_ = obj.get_field_by_name(id_attribute)
            if id_ not in ids_seen:
                ids_seen.add(id_)
                yield str(id_)  # May need to revisit this string conversion

    @classmethod
    def get_objects_from_ids(
        cls,
        datasource: DataSource,
        object_type: str,
        ids: Iterator[str],
        sort_by: str = None,
    ) -> Iterator[DataObject]:
        objs = datasource.get_by_ids(object_type, ids)
        if sort_by is not None:
            yield from sorted(objs, key=lambda obj: obj.get_field_by_name(sort_by) or 0)
        else:
            yield from objs



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..core import (
    core_data_object
)
from ..jira import (
    JiraDataSource,
    create_jira_datasource
)


def grit(**kwargs) -> JiraDataSource:
    grit = create_jira_datasource(
        jira_url=os.getenv('JIRA_URL', Defaults.JIRA_URL),
        jira_api_key=os.getenv('JIRA_API_KEY')
    )
    core_data_object(grit)
    return grit



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .abc import AuthManager  # noqa F401
from .asserts import AuthInspector, basic_auth_inspector, require_auth  # noqa F401
from .blueprint import AuthBlueprint  # noqa F401
from .composite import CompositeAuthInspector  # noqa F401
from .config import OidcConfig, env_oidc_config  # noqa F401
from .error import AuthError, ForbiddenError, StateNotFoundError  # noqa F401



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Any

import dateparser

from ..datasource_filter import DataSourceFilter


class _Filterable(ABC):
    """
    Implements methods for data sources that support filtering
    e.g. ListGetter, PageGetter and Counter
    """

    def _preprocess_filter(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None = None,
    ) -> DataSourceFilter | None:
        """
        This method is called inside a datasource before starting its associated filter converter.
        This method, unlike the converter, converts in place.
        By default, this method converts relative dates into absolute dates.
        """
        if object_filters is None:
            return None
        if object_filters.and_ is not None:
            for name, value in object_filters.and_.items():
                metadata = self.get_attribute_metadata_by_name(object_type, name)
                if metadata is None:
                    continue
                for op, val in value.items():
                    if 'value' in val and metadata['python_type'] == 'datetime' \
                            and isinstance(val['value'], str):
                        object_filters.and_[name][op]['value'] = dateparser.parse(val['value'])

        return object_filters

    @abstractmethod
    def get_attribute_metadata_by_name(self, obj_type: str, field_name: str) -> Any:
        raise NotImplementedError('Should be implemented by DataSource')



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TreeofsexSpeciesToTreeofsexwhSpeciesConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        attributes = {
            att.attribute_key.id: att.value
            for att in data_object.atts
        }

        ret = self._data_object_factory(
            'species',
            data_object.id,
            attributes=attributes,
        )
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class GritIssueToElasticCurationConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        assembly_stats = self.__get_assembly_stats(
            data_object.attributes.get('assembly_statistics')
        )
        chr_data = self.__get_chr_data(data_object.attributes.get('chromosome_result'))
        attributes = {
            k: v for k, v in data_object.attributes.items()
            if k not in ['assembly_statistics', 'chromosome_result', 'description',
                         'sample_id', 'status_changes', 'linked_issues']
        } | {
            self.__sanitise_attribute_name(sc['next_status']) + '_date': sc['end_date']
            for sc in data_object.status_changes
        } | {
            'assignee_name': data_object.assignee.name if data_object.assignee else None,
        } | assembly_stats | chr_data

        to_one_relations = {
            'tolid': self._data_object_factory(
                'tolid',
                data_object.attributes.get('sample_id')
            )
        }
        ret = self._data_object_factory(
            'curation',
            data_object.id,
            attributes=attributes,
            to_one=to_one_relations
        )
        yield ret

    def __sanitise_attribute_name(self, name: str) -> str:
        return re.sub(r'\s+', '_', name.lower())

    def __get_assembly_stats(self, data):
        if not data:
            return {}
        assembly_stats = {}
        pattern = re.compile(
            r'(?P<section>scaffolds|contigs)\n(?P<section_data>'
            r'(?:[a-zA-Z0-9]+\s+\d+\s+\d+\s*\n?)+)'
        )
        for match in pattern.finditer(data):
            section = match.group('section')
            section_data = match.group('section_data')
            for att in ['total', 'count', 'N50', 'L50', 'N90', 'L90']:
                assembly_stats.update(self.__get_assembly_info(section_data, section, att))
        return assembly_stats

    def __get_assembly_info(self, data, contig_or_scaffold, att):
        """
        Function to return the information hidden in assembly stats
        """
        if data:
            att_search = re.search(rf'{att}\s*([0-9]\w*)\s*([0-9]\w*)', data)
            att_before = int(att_search.group(1))
            att_after = int(att_search.group(2))
            att_change_per = (att_after - att_before) / att_before * 100
            return {
                f'{contig_or_scaffold}_{att.lower()}_before': att_before,
                f'{contig_or_scaffold}_{att.lower()}_after': att_after,
                f'{contig_or_scaffold}_{att.lower()}_change_per': att_change_per
            }
        else:
            return {}

    def __get_chr_data(self, chromo_res):
        """
        Function to parse and return the chromosome assignment and assignment %
        :param chromo_res:
        :return:
        """
        if chromo_res:
            chr_ass_search = re.search(r'(found.[0-9].*somes.*\(.*\))', chromo_res)
            if chr_ass_search:
                chr_ass = chr_ass_search.group(1)
            elif chr_ass_search is None:
                chr_ass_search = re.search(r'(found.[0-9].*somes)', chromo_res)
                if chr_ass_search:
                    chr_ass = chr_ass_search.group(1)
                else:
                    chr_ass = None
            else:
                chr_ass = None

            ass_percent_search = re.search(r'Chr.length.(\d*.\d*).%', chromo_res)
            ass_percent = ass_percent_search.group(1) if ass_percent_search else None
            return {
                'chr_ass': chr_ass,
                'ass_percent': ass_percent
            }
        else:
            return {}



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from .attribute_metadata import *  # noqa F401
from .core_converter import (  # noqa F401
    ChainedConverter,
    Converter,
    is_iter
)
from .data_source_dict import DataSourceDict  # noqa F401
from .datasource import *  # noqa F401
from .datasource_error import DataSourceError  # noqa F401
from .datasource_filter import DataSourceFilter  # noqa F401
from .datasource_filter_converter import DataSourceFilterConverter  # noqa F401
from .datasource_utils import DataSourceUtils  # noqa F401
from .data_loader import (  # noqa F401
    DataLoader,
    DefaultDataLoader,
    GroupStatterDataLoader,
    IdsDataLoader,
    ObjectsDataLoader
)
from .data_object import (  # noqa F401
    DataObject,
    ErrorObject
)
from .data_object_converter import (  # noqa F401
    DataObjectToDataObjectOrUpdateConverter,
    DefaultDataObjectToDataObjectConverter,
    SanitisingConverter
)
from .factory import core_data_object  # noqa F401
from .http_client import HttpClient  # noqa F401
from .requested_fields import ReqFieldsTree  # noqa: F401
from .validate import (  # noqa F401
    Validator,
    ValidationSeverity,
    ValidationResult
)
from .datasource_parser import DataSourceParser  # noqa F401



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import (
    Iterable
)

from .elastic_sample_to_benchling_tissue_update_converter import (
    ElasticSampleToBenchlingTissueUpdateConverter
)
from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class ElasticSampleToBenchlingTissueConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def __init__(self, data_object_factory):
        super().__init__(data_object_factory)
        self.update_converter = ElasticSampleToBenchlingTissueUpdateConverter(
            data_object_factory=self._data_object_factory
        )

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        converted_update = self.update_converter._convert_one(data_object)
        if converted_update is not None:
            id_, attributes = converted_update
            ret = self._data_object_factory(
                'tissue',
                id_,
                attributes=attributes
            )
            yield ret
        else:
            yield None



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

"""
Controller module for MVC-style request handling.

This module provides the Controller class which serves as an intermediary between
data sources and views, handling various operations such as CRUD operations,
aggregations, and relationship management with proper validation and authorisation.
"""

from __future__ import annotations

import importlib
import inspect
from datetime import datetime
from inspect import BoundArguments
from typing import Any, Callable, Iterable, Optional, Type


from .auth import AuthInspector
from .misc import (
    ActionParameters,
    AggregationBody,
    AggregationParameters,
    CtxGetter,
    GroupStatsParameters,
    ListGetParameters,
    StatsParameters,
    default_ctx_getter,
)
from ..api_client.exception import (
    ObjectNotFoundByIdException,
    RecursiveRelationNotFoundException,
    UninheritedOperationError,
    UnsupportedOperationError,
)
from ..api_client.view import ResponseDict, View
from ..core import DataObject, DataSourceError, OperableDataSource, ReqFieldsTree
from ..core.datasource_filter import AndFilter, DataSourceFilter
from ..core.operator import (
    Aggregator,
    Counter,
    Cursor,
    Deleter,
    DetailGetter,
    GroupStatter,
    Inserter,
    Operator,
    OperatorMethod,
    PageGetter,
    Relational,
    ReturnMode,
    Updater,
    Upserter,
)
from ..core.operator.updater import DataObjectUpdate


EmptySuccessResponse = dict[str, bool]


def __is_supported(
    operator_class: Type[Operator],
    operator_method: str,
    data_source: OperableDataSource,
) -> bool:
    """
    Check if a given DataSource instance supports the specified Operator class.

    This function determines whether the data source implements the required operator
    by checking instance relationships and method availability. It handles inheritance
    validation to ensure proper operator implementation.

    Args:
        operator_class: The Operator class type to check for support.
        operator_method: The name of the operator method to validate.
        data_source: The OperableDataSource instance to examine.

    Returns:
        True if the data source supports the operator class, False otherwise.

    Raises:
        UninheritedOperationError: If the operator method exists on the data source
            but the required operator class is not properly inherited.
    """
    if isinstance(data_source, operator_class):
        return True
    if hasattr(data_source, operator_method):
        raise UninheritedOperationError(data_source, operator_class, operator_method)
    return False


def validate(
    operator_class: Type[Operator],
    object_method_name: str,
    operator_method: OperatorMethod,
) -> Callable:
    """
    Decorator factory for validating Controller method operations.

    This decorator ensures that a Controller method's corresponding operation is
    supported by its DataSource. It performs comprehensive validation including
    inheritance checking and authorisation inspection.

    Args:
        operator_class: The Operator class that must be supported by the data source.
        object_method_name: The name of the method on the data source object.
        operator_method: The OperatorMethod enum value for this operation.

    Returns:
        A decorator function that validates the operation before execution.

    Raises:
        UninheritedOperationError: If the method is implemented but the mixin ABC
            is not inherited from.
        UnsupportedOperationError: If the operation is not supported by the data source.
    """

    def decorator(method: Callable) -> Callable:
        sig = inspect.signature(method)

        def wrapper(controller: Controller, object_type: str, *args, **kwargs) -> Any:
            bound_args = sig.bind(controller, object_type, *args, **kwargs)
            bound_args.apply_defaults()

            if not __is_supported(
                operator_class, object_method_name, controller.data_source
            ):
                raise UnsupportedOperationError(object_type, str(operator_method))

            ext_and = controller.inspect_auth(object_type, operator_method, bound_args)
            return method(
                controller,
                object_type,
                *args,
                ext_and=ext_and,
                **kwargs,
            )

        return wrapper

    return decorator


class Controller:
    """
    MVC-style Controller class for handling API requests and coordinating data operations.

    The Controller serves as an intermediary between the data source layer and the view
    layer, providing a consistent interface for various operations including CRUD operations,
    aggregations, statistics, and relationship management. It handles validation,
    authorisation, and error management for all supported operations.

    The controller supports a wide range of operations:
    - Detail retrieval and listing with pagination
    - Counting and statistics generation
    - Create, Read, Update, Delete (CRUD) operations
    - Aggregation queries
    - Cursor-based pagination
    - Relationship traversal (to-one and to-many)
    """

    def __init__(
        self,
        data_source: OperableDataSource,
        view: View,
        requested_tree: ReqFieldsTree | None = None,
        auth_inspector: Optional[AuthInspector] = None,
    ) -> None:
        """
        Initialise the Controller with required dependencies.

        Args:
            data_source: The OperableDataSource instance for data operations.
            view: The View instance for response formatting.
            auth_inspector: Optional AuthInspector for authorisation validation.
                If None, no authorisation checks will be performed.
        """
        self.__data_source = data_source
        self.__view = view
        self.__requested_tree = requested_tree
        self.__inspector = auth_inspector

    @property
    def data_source(self) -> OperableDataSource:
        """
        Get the configured data source instance.

        Returns:
            The OperableDataSource instance used by this controller.
        """
        return self.__data_source

    def inspect_auth(
        self, object_type: str, operation: OperatorMethod, bound_args: BoundArguments
    ) -> Optional[AndFilter]:
        """
        Perform authorisation inspection for the given operation.

        This method delegates to the configured AuthInspector to determine if the
        current context has permission to perform the specified operation on the
        given object type.

        Args:
            object_type: The type of object being operated on.
            operation: The OperatorMethod being performed.
            bound_args: The bound arguments from the calling method.

        Returns:
            An optional AndFilter containing authorisation constraints, or None
            if no inspector is configured or no constraints are required.
        """
        if self.__inspector is not None:
            return self.__inspector(object_type, operation, bound_args)

    @validate(DetailGetter, 'get_by_id', OperatorMethod.DETAIL)
    def get_detail(self, object_type: str, object_id: str, **kwargs) -> ResponseDict:
        """
        Retrieve an individual object of the specified type and identifier.

        This method fetches a single data object by its unique identifier and
        returns it in a format suitable for API responses.

        Args:
            object_type: The type of object to retrieve.
            object_id: The unique identifier of the object.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            A ResponseDict containing the serialised object data.

        Raises:
            ObjectNotFoundByIdException: If no object with the given ID exists.
            UnsupportedOperationError: If the data source doesn't support detail retrieval.
        """
        data_object = self.__get_detail_object(object_type, object_id)
        return self.__view.dump(data_object)

    @validate(PageGetter, 'get_list_page', OperatorMethod.PAGE)
    def get_list(
        self,
        object_type: str,
        query_args: ListGetParameters,
        ext_and: Optional[AndFilter] = None,
    ) -> ResponseDict:
        """
        Retrieve a paginated list of objects of the specified type.

        This method fetches a page of objects based on the provided query parameters,
        including filters, sorting, and field selection. It returns both the data
        and metadata about the total count and available types.

        Args:
            object_type: The type of objects to retrieve.
            query_args: Parameters controlling pagination, filtering, and sorting.
            ext_and: Optional additional filter constraints from authorisation.

        Returns:
            A ResponseDict containing the paginated objects and metadata including
            total count and attribute type information.

        Raises:
            UnsupportedOperationError: If the data source doesn't support page retrieval.
        """
        page_number = self.__get_page_number_or_1(query_args)
        data_objects, total = self.__data_source.get_list_page(
            object_type,
            page_number,
            page_size=query_args.page_size,
            object_filters=self.__combine_filters(query_args.filter, ext_and),
            sort_by=query_args.sort_by,
            requested_tree=self.__requested_tree,
        )
        meta = {
            'total': total,
            'types': self.__data_source.get_attribute_types(object_type),
        }
        return self.__view.dump_bulk(data_objects, document_meta=meta)

    @validate(Counter, 'get_count', OperatorMethod.COUNT)
    def get_count(
        self,
        object_type: str,
        query_args: ListGetParameters,
        ext_and: Optional[AndFilter] = None,
    ) -> ResponseDict:
        """
        Get the count of objects of the specified type matching the given filters.

        This method returns only the total count of objects without retrieving
        the actual data, which is more efficient for counting operations.

        Args:
            object_type: The type of objects to count.
            query_args: Parameters containing filter conditions.
            ext_and: Optional additional filter constraints from authorisation.

        Returns:
            A ResponseDict containing metadata with the total count.

        Raises:
            UnsupportedOperationError: If the data source doesn't support counting.
        """
        total = self.__data_source.get_count(
            object_type,
            object_filters=self.__combine_filters(query_args.filter, ext_and),
        )
        document_meta = {'total': total}
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(Counter, 'get_stats', OperatorMethod.STATS)
    def get_stats(
        self,
        object_type: str,
        query_args: StatsParameters,
        ext_and: Optional[AndFilter] = None,
    ) -> ResponseDict:
        """
        Generate statistical information for objects of the specified type.

        This method calculates various statistics (such as min, max, average, sum)
        for specified fields on objects matching the given filters.

        Args:
            object_type: The type of objects to analyse.
            query_args: Parameters specifying which statistics to calculate and on which fields.
            ext_and: Optional additional filter constraints from authorisation.

        Returns:
            A ResponseDict containing metadata with the calculated statistics.

        Raises:
            UnsupportedOperationError: If the data source doesn't support statistics.
        """
        stats = self.__data_source.get_stats(
            object_type,
            stats=query_args.stats,
            stats_fields=query_args.stats_fields,
            object_filters=self.__combine_filters(query_args.filter, ext_and),
        )
        document_meta = {**stats, 'type': object_type}
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(GroupStatter, 'get_group_stats', OperatorMethod.GROUP_STATS)
    def get_group_stats(
        self,
        object_type: str,
        query_args: GroupStatsParameters,
        ext_and: Optional[AndFilter] = None,
    ) -> ResponseDict:
        """
        Generate grouped statistical information for objects of the specified type.

        This method calculates statistics grouped by one or more fields, allowing
        for analysis of data distribution across different categories.

        Args:
            object_type: The type of objects to analyse.
            query_args: Parameters specifying grouping fields, statistics, and target fields.
            ext_and: Optional additional filter constraints from authorisation.

        Returns:
            A ResponseDict containing metadata with the grouped statistics.

        Raises:
            UnsupportedOperationError: If the data source doesn't support group statistics.
        """
        stats = self.__data_source.get_group_stats(
            object_type,
            query_args.group_by,
            stats=query_args.stats,
            stats_fields=query_args.stats_fields,
            object_filters=self.__combine_filters(query_args.filter, ext_and),
        )
        document_meta = {'stats': list(stats), 'type': object_type}
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(Deleter, 'delete', OperatorMethod.DELETE)
    def delete_detail(
        self, object_type: str, object_id: str, **kwargs
    ) -> EmptySuccessResponse:
        """
        Delete the DataObject of the specified type and identifier.

        This method removes a single object from the data source. The operation
        is permanent and cannot be undone.

        Args:
            object_type: The type of object to delete.
            object_id: The unique identifier of the object to delete.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            An EmptySuccessResponse indicating successful deletion.

        Raises:
            UnsupportedOperationError: If the data source doesn't support deletion.
        """
        self.data_source.delete(object_type, [object_id])
        return {'success': True}

    @validate(Updater, 'update', OperatorMethod.UPDATE)
    def patch_list(
        self, object_type: str, updates: Iterable[DataObjectUpdate], **kwargs
    ) -> EmptySuccessResponse:
        """
        Update multiple objects of the same type using the provided update specifications.

        This method applies updates to existing objects based on their identifiers
        and the update data provided. All objects must be of the same type.

        Args:
            object_type: The type of objects to update.
            updates: An iterable of DataObjectUpdate instances containing
                ID and update data pairs.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            An EmptySuccessResponse indicating a successful update.

        Raises:
            UnsupportedOperationError: If the data source doesn't support updates.
        """
        self.data_source.update(object_type, updates)
        return {'success': True}

    @validate(Inserter, 'insert', OperatorMethod.INSERT)
    def post_inserts(
        self, object_type: str, objects: Iterable[DataObject], **kwargs
    ) -> EmptySuccessResponse:
        """
        Insert new objects of the specified type into the data source.

        This method creates new objects in the data source. The return value
        depends on the data source's return mode configuration.

        Args:
            object_type: The type of objects to insert.
            objects: An iterable of DataObject instances to insert.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            Either an EmptySuccessResponse or a ResponseDict containing the inserted
            objects, depending on the data source's return mode.

        Raises:
            UnsupportedOperationError: If the data source doesn't support insertion.
        """
        returned = self.data_source.insert(object_type, objects)
        if self.data_source.return_mode[object_type] == ReturnMode.POPULATED:
            return self.__view.dump_bulk(returned)
        else:
            return {'success': True}

    @validate(Upserter, 'post_upserts', OperatorMethod.UPSERT)
    def post_upserts(
        self, object_type: str, objects: Iterable[DataObject], **kwargs
    ) -> EmptySuccessResponse:
        """
        Upsert (insert or update) objects of the specified type.

        This method performs an upsert operation, which will insert new objects
        or update existing ones based on their identifiers. The return value
        depends on the data source's return mode configuration.

        Args:
            object_type: The type of objects to upsert.
            objects: An iterable of DataObject instances to upsert.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            Either an EmptySuccessResponse or a ResponseDict containing the upserted
            objects, depending on the data source's return mode.

        Raises:
            UnsupportedOperationError: If the data source doesn't support upsertion.
        """
        returned = self.data_source.upsert(object_type, objects, **kwargs)
        if self.data_source.return_mode[object_type] == ReturnMode.POPULATED:
            return self.__view.dump_bulk(returned)
        else:
            return {'success': True}

    @validate(Aggregator, 'get_aggregations', OperatorMethod.AGGREGATE)
    def post_aggregations(
        self,
        object_type: str,
        query_args: AggregationParameters,
        body: AggregationBody,
        ext_and: Optional[AndFilter] = None,
    ) -> ResponseDict:
        """
        Perform aggregation operations on objects of the specified type.

        This method executes complex aggregation queries such as grouping,
        bucketing, and statistical calculations on the data set.

        Args:
            object_type: The type of objects to aggregate.
            query_args: Parameters containing filter conditions.
            body: The aggregation specification defining the operations to perform.
            ext_and: Optional additional filter constraints from authorisation.

        Returns:
            A ResponseDict containing the aggregation results and type information.

        Raises:
            UnsupportedOperationError: If the data source doesn't support aggregation.
        """
        aggregation_results = self.__data_source.get_aggregations(
            object_type,
            object_filters=self.__combine_filters(query_args.filter, ext_and),
            aggregations=body.aggs,
        )
        document_meta = {
            'aggregations': aggregation_results,
            'types': self.__data_source.get_attribute_types(object_type),
        }
        return self.__view.dump_bulk([], document_meta=document_meta)

    @validate(Cursor, 'get_cursor_page', OperatorMethod.CURSOR)
    def get_cursor_page(
        self,
        object_type: str,
        query_args: ListGetParameters,
        search_after: list[str] | None,
        ext_and: Optional[AndFilter] = None,
    ) -> ResponseDict:
        """
        Retrieve a page of objects using cursor-based pagination.

        This method provides an alternative pagination mechanism using cursors,
        which is more efficient for large datasets and provides consistent
        results even when the underlying data changes.

        Args:
            object_type: The type of objects to retrieve.
            query_args: Parameters containing filter conditions and page size.
            search_after: The cursor position to start from, or None for the first page.
            ext_and: Optional additional filter constraints from authorisation.

        Returns:
            A ResponseDict containing the objects and metadata, including the
            new cursor position for the next page.

        Raises:
            UnsupportedOperationError: If the data source doesn't support cursor pagination.
        """
        data_objects, new_search_after = self.data_source.get_cursor_page(
            object_type,
            query_args.page_size,
            self.__combine_filters(query_args.filter, ext_and),
            search_after,
            requested_tree=self.__requested_tree,
        )
        meta = {'search_after': new_search_after}

        return self.__view.dump_bulk(data_objects, document_meta=meta)

    @validate(Relational, 'get_recursive_relation', OperatorMethod.TO_ONE)
    def get_recursive_relation(
        self, data_object: DataObject, relationship_hops: list[str], **kwargs
    ) -> ResponseDict:
        """
        Retrieve a nested to-one relationship by following multiple relationship hops.

        This method allows traversal of complex object relationships by following
        a chain of to-one relationships defined by the relationship_hops parameter.

        Args:
            data_object: The source DataObject to start the traversal from.
            relationship_hops: A list of relationship names to traverse in order.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            A ResponseDict containing the final related object.

        Raises:
            RecursiveRelationNotFoundException: If the relationship chain cannot
                be followed or the final object is None.
            UnsupportedOperationError: If the data source doesn't support relationships.
        """
        related_object = self.__get_to_one_relation(data_object, relationship_hops)
        if related_object is None:
            raise RecursiveRelationNotFoundException()
        return self.__view.dump(related_object)

    @validate(Relational, 'get_to_many_relations_page', OperatorMethod.TO_MANY)
    def get_many_relations_page(
        self,
        data_object: DataObject,
        relationship_name: str,
        query_args: ListGetParameters,
        **kwargs,
    ) -> ResponseDict:
        """
        Retrieve a paginated list of objects from a to-many relationship.

        This method fetches related objects that are connected through a
        to-many relationship, with support for pagination.

        Args:
            data_object: The source DataObject containing the relationship.
            relationship_name: The name of the to-many relationship to traverse.
            query_args: Parameters controlling pagination.
            **kwargs: Additional keyword arguments (including ext_and from validation).

        Returns:
            A ResponseDict containing the paginated related objects.

        Raises:
            UnsupportedOperationError: If the data source doesn't support relationships.
        """
        page = self.data_source.get_to_many_relations_page(
            data_object, relationship_name, query_args.page, query_args.page_size
        )
        return self.__view.dump_bulk(page)

    def __combine_filters(
        self,
        object_filters: Optional[DataSourceFilter] = None,
        ext_and: Optional[AndFilter] = None,
    ) -> Optional[DataSourceFilter]:
        """
        Combine user-provided filters with authorisation filters.

        This private method merges filters from query parameters with additional
        filters imposed by the authorisation system, ensuring that both sets
        of constraints are properly applied.

        Args:
            object_filters: Optional filters from the user query.
            ext_and: Optional additional AND filters from authorisation.

        Returns:
            A combined DataSourceFilter or None if no filters are present.
        """
        if ext_and is None:
            return object_filters
        else:
            if object_filters is None:
                return DataSourceFilter(and_=ext_and)
            elif object_filters.and_ is None:
                object_filters.and_ = ext_and
            else:
                object_filters.and_ |= ext_and

            return object_filters

    def __get_detail_object(self, object_type: str, object_id: str) -> DataObject:
        """
        Retrieve a single DataObject by type and identifier.

        This private method fetches an object and handles the case where
        the object doesn't exist or is None.

        Args:
            object_type: The type of object to retrieve.
            object_id: The unique identifier of the object.

        Returns:
            The requested DataObject.

        Raises:
            ObjectNotFoundByIdException: If no object with the given ID exists.
        """
        data_objects = list(
            self.__data_source.get_by_id(
                object_type,
                [object_id],
                requested_tree=self.__requested_tree,
            )
        )
        if len(data_objects) == 0 or data_objects[0] is None:
            raise ObjectNotFoundByIdException(object_type, object_id)
        return data_objects[0]

    def __get_page_number_or_1(self, query_args: ListGetParameters) -> int:
        """
        Extract the page number from query arguments or return 1 as default.

        This private method provides a safe way to get the page number,
        defaulting to page 1 if no page is specified.

        Args:
            query_args: The query parameters containing the optional page number.

        Returns:
            The page number to use, defaulting to 1.
        """
        page_number = query_args.page
        if page_number is None:
            return 1
        return page_number

    def __get_to_one_relation(
        self, source: DataObject, relationship_hops: list[str]
    ) -> Optional[DataObject]:
        """
        Traverse a to-one relationship chain and return the final related object.

        This private method validates the relationship chain and performs
        the actual traversal through the data source.

        Args:
            source: The source DataObject to start traversal from.
            relationship_hops: The list of relationship names to traverse.

        Returns:
            The final related DataObject, or None if the chain cannot be followed.
        """
        self.__data_source.validate_to_one_recurse(source.type, relationship_hops)
        return self.__data_source.get_recursive_relation(source, relationship_hops)

    def perform_action(
        self,
        object_type: str,
        action_args: ActionParameters,
        action_ds: OperableDataSource,
        flow_ds: OperableDataSource,
    ) -> ResponseDict:
        """
        Perform an action on objects of the specified type.

        This method executes a predefined action based on the provided parameters,
        which may include an action name, target IDs, and additional parameters.

        Args:
            object_type: The type of objects to perform the action on.
            action_name: The name of the action to perform.
            ids: List of object IDs to target with the action.
            params: Optional additional parameters for the action.

        Returns:
            A dictionary containing the result of the action.
        """
        # Implementation of the action execution logic goes here.
        ctx: CtxGetter = default_ctx_getter()

        user_id = ctx.user_id

        # TODO: Add role-based access control for actions here, potentially using ctx.roles

        ids: list[str] = action_args.ids
        action_name: str = action_args.action_name
        params: dict[str, Any] = action_args.params

        action = self.__get_action(object_type, action_name, action_ds)
        action_roles = self.__get_role_action_list(action, action_ds)

        if len(action_roles) == 0:
            raise DataSourceError(
                'Unauthorized',
                'This action does not have any roles assigned to it',
                403
            )

        for action_role in action_roles:
            if action_role not in ctx.roles:
                raise DataSourceError(
                    'Unauthorized',
                    'User does not have required role for this action',
                    403
                )

        action_params = (
            action.params
            if action.params
            else {}
        )

        if action.flow_name:
            flow_params = {
                'extra_params': {
                    **params,
                    **action_params,
                },
                'user_id': user_id,
                'object_type': object_type,
                'ids': ids
            }

            flow_run_id, flow_run_name = self.__insert_flow_run(
                action,
                flow_params,
                user_id,
                flow_ds
            )

            user_action_params = {
                **params,
                **action_params,
                'ids': ids,
                'flow_run_id': flow_run_id,
                'flow_run_name': flow_run_name
            }

        elif action.class_name:
            # Try to import the class from tol.actions first, then fall back to main.actions
            action_class = None
            try:
                tol_actions_module = importlib.import_module('tol.actions')
                if hasattr(tol_actions_module, action.class_name):
                    action_class = getattr(tol_actions_module, action.class_name)

                if action_class is None:
                    main_actions_module = importlib.import_module('main.actions')
                    if hasattr(main_actions_module, action.class_name):
                        action_class = getattr(main_actions_module, action.class_name)

            except ImportError:
                raise DataSourceError(
                    'Action Class Import Error',
                    'Class not found in tol.actions or main.actions',
                    500
                )

            if action_class is None:
                raise DataSourceError(
                    'Action Class Not Found',
                    f'Action class "{action.class_name}" not found in tol.actions or main.actions',
                    404
                )

            class_params = {**action_params, **params}

            action_instance = action_class()
            status = action_instance.run(ids=ids, params=class_params,
                                         object_type=object_type, datasource=action_ds)

            user_action_params = {
                **params,
                **action_params,
                'ids': ids,
                'status': status
            }

        else:
            raise DataSourceError(
                'Invalid Action',
                'No Actions are defined',
                400
            )

        user = action_ds.get_one('user', user_id)

        user_action = action_ds.data_object_factory(
            'user_action',
            attributes={
                'ids': ids,
                'params': user_action_params,
                'created_at': datetime.now()
            },
            to_one={
                'user': user,
                'action': action
            }
        )
        action_ds.insert('user_action', [user_action])

        return {'success': True}, 200

    def __get_action(
        self,
        object_type: str,
        action_name: str,
        action_ds: OperableDataSource
    ) -> DataObject:

        f = DataSourceFilter(
            and_={
                'name': {
                    'eq': {
                        'value': action_name
                    }
                },
                'object_type': {
                    'eq': {
                        'value': object_type
                    }
                }
            }
        )

        action_list = list(
            action_ds.get_list(
                'action',
                object_filters=f
            )
        )

        if not action_list:
            raise DataSourceError(
                'Not Found',
                'The specified action was not found',
                404
            )

        return action_list[0]

    def __insert_flow_run(
        self,
        action: DataObject,
        flow_params: dict[str, Any],
        user_id: str,
        flow_ds: OperableDataSource
    ) -> tuple[str, str]:

        flow_name = action.flow_name
        flow_run = flow_ds.data_object_factory(
            'flow_run',
            attributes={
                'flow_name': flow_name,
                'deployment_name': flow_name,
                'parameters': flow_params,
                'tags': [
                    'app_name:portal',
                    f'user_id:{user_id}'
                ],
            }
        )

        inserted_run_data = list(
            flow_ds.insert(
                'flow_run',
                [flow_run]
            )
        )[0]

        return inserted_run_data.id, inserted_run_data.name

    def __get_role_action_list(
        self,
        action: DataObject,
        action_ds: OperableDataSource
    ) -> list[str]:
        role_action_list = action_ds.get_list(
            'role_action',
            object_filters=DataSourceFilter(
                and_={
                    'action_id': {
                        'eq': {
                            'value': action.id
                        }
                    }
                }
            )
        )
        role_id_list = []
        for role_action in role_action_list:
            role_id = role_action.to_one_relationships['role'].id
            role_id_list.append(role_id)

        return self.__get_roles_from_id(role_id_list, action_ds)

    def __get_roles_from_id(
        self,
        role_ids: list[str],
        action_ds: OperableDataSource
    ) -> list[str]:
        role_list = action_ds.get_list(
            'role',
            object_filters=DataSourceFilter(
                and_={
                    'id': {
                        'in_list': {
                            'value': role_ids
                        }
                    }
                }
            )
        )
        return [role.name for role in role_list]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import (Dict, Iterable, List)

import requests

from ..core import (
    DataObject,
    DataSource,
    DataSourceError
)
from ..core.operator import (
    Upserter
)


class StsDataSource(DataSource, Upserter):
    """
    This is a very simple DataSource that provides access to certain writeable
    endpoints in STS via the standard DataSource methods (upsert, etc.)
    There is also direct access to any endpoint via the native_***() methods
    """

    def __init__(self, config: Dict):
        super().__init__(config, expected=['url', 'key'])

    @property
    def supported_types(self) -> List[str]:
        return ['sequencing_request', 'extraction']

    @property
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'sequencing_request': {
                'platform': 'str',
                'fluidx_id': 'str',
                'submission_date': 'datetime'
            },
            'extraction': {
                'sample_id': 'str',
                'fluidx_id': 'str',
                'extraction_type': 'str',
                'extraction_date': 'datetime'
            }
        }

    def _encode_date(self, d):
        return d.strftime('%Y-%m-%d %H:%M:%S')

    def __post_sequencing_requests(self, sequencing_requests: Iterable[DataObject]):
        """
        We are expecting a list of DataObjects with:
        id
        platform
        fluidx_id
        submission_date
        """
        for sr in sequencing_requests:
            payload = {'platform': sr.platform,
                       'fluidx_id': sr.fluidx_id,
                       'sample_ref': sr.id,
                       'submit_date': self._encode_date(sr.submission_date)}
            r = self.native_post(
                '/sequencing-requests',
                json=payload
            )
            if r.ok:
                print(f'OK: {sr.id}')
            else:
                print(
                    f'A sample failed with code {r.status_code}, '
                    f'and response {r.json()}, '
                    f'containing data: {payload}'
                )

    def __post_extractions(self, extractions: Iterable[DataObject]):
        """
        We are expecting a list of DataObjects with:
        id (the ELN id)
        fluidx_id
        sample_id
        extraction_type (uppercase)
        extraction_date
        """
        for extraction in extractions:
            payload = {'eln_id': extraction.id,
                       'sample_id': extraction.sample_id,
                       'fluidx_id': extraction.fluidx_id,
                       'type': extraction.extraction_type,
                       'extraction_date': self._encode_date(extraction.extraction_date)}
            r = self.native_post(
                f'/ep_samples/{extraction.fluidx_id}',
                json=payload
            )
            if r.ok:
                print(f'OK: {extraction.id}')
            else:
                print(
                    f'A sample failed with code {r.status_code}, '
                    f'and response {r.json()}, '
                    f'containing data: {payload}'
                )

    def upsert(
        self,
        object_type: str,
        objects: Iterable[DataObject]
    ) -> None:
        if object_type == 'sequencing_request':
            self.__post_sequencing_requests(objects)
        elif object_type == 'extraction':
            self.__post_extractions(objects)
        else:
            raise DataSourceError('Only objects of type sequencing_request and extraction '
                                  'are handled by StsDataSource')

    # Everything below here is to do with allowing native endpoint requests

    def __override_method(self, method, relative_url, headers=None, **kwargs):
        if headers is None:
            new_headers = {
                'Authorization': self.key
            }
        else:
            new_headers = {
                'Authorization': self.key,
                **headers
            }
        return method(
            f'{self.url}/{relative_url}',
            headers=new_headers,
            **kwargs
        )

    def native_get(self, relative_url, **kwargs):
        return self.__override_method(
            requests.get,
            relative_url,
            **kwargs
        )

    def native_post(self, relative_url, **kwargs):
        return self.__override_method(
            requests.post,
            relative_url,
            **kwargs
        )

    def native_put(self, relative_url, **kwargs):
        return self.__override_method(
            requests.put,
            relative_url,
            **kwargs
        )

    def native_patch(self, relative_url, **kwargs):
        return self.__override_method(
            requests.patch,
            relative_url,
            **kwargs
        )

    def native_delete(self, relative_url, **kwargs):
        return self.__override_method(
            requests.delete,
            relative_url,
            **kwargs
        )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..core import (
    core_data_object
)
from ..labwhere import (
    LabwhereDataSource,
    create_labwhere_datasource
)


def labwhere(**kwargs) -> LabwhereDataSource:
    labwhere = create_labwhere_datasource(
        labwhere_url=os.getenv('LABWHERE_URL', Defaults.LABWHERE_URL)
        + os.getenv('LABWHERE_API_PATH', Defaults.LABWHERE_API_PATH)
    )
    core_data_object(labwhere)
    return labwhere



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import asyncio
from typing import Any, Callable, Iterable, Optional, Union
from uuid import UUID

from prefect.client.orchestration import PrefectClient
from prefect.client.schemas.objects import FlowRun
from prefect.client.schemas.sorting import FlowRunSort
from prefect.exceptions import ObjectNotFound

from .converter import ObjectConverter, PrefectConverter
from .filter import PrefectFilter
from .prefect_object import FlowRunObject
from ..core import DataSource, DataSourceError, DataSourceFilter
from ..core.operator import (
    DetailGetter,
    Inserter,
    ListGetter,
    PageGetter
)


class PrefectDataSource(
    DataSource,

    DetailGetter,
    Inserter,
    ListGetter,
    PageGetter
):
    """
    Manages the prefect async infrastructure.

    Most developers will wish to use
    `create_prefect_datasource`, instead of this
    directly.
    """

    def __init__(
        self,
        client_factory: Callable[[], PrefectClient],
        filter_factory: Callable[[], PrefectFilter],
        object_converter_factory: Callable[[], ObjectConverter],
        prefect_converter_factory: Callable[[], PrefectConverter]
    ) -> None:

        self.__client_factory = client_factory
        self.__filter_factory = filter_factory
        self.__oc_factory = object_converter_factory
        self.__prc_factory = prefect_converter_factory

        super().__init__({})

    @property
    def supported_types(self) -> list[str]:
        return ['flow_run']

    @property
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'flow_run': {
                'flow_name': 'str',
                'deployment_name': 'str',
                'name': 'str',
                'tags': 'list[str]',
                'state': 'str',
                'idempotency_key': 'str',
                'parameters': 'dict[str, Any]'
            }
        }

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[Optional[FlowRunObject]]:

        self.__validate_object_type(object_type)

        return asyncio.run(
            self.__async_get_by_id(object_ids)
        )

    def get_flow_run(self, flow_run_id: str) -> Optional[FlowRunObject]:
        """
        Gets the `FlowRunObject` with the given `flow_run_id`.

        Returns `None` if not found.
        """

        iter_results = self.get_by_id('flow_run', [flow_run_id])
        return iter_results[0]

    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None
    ) -> tuple[Iterable[FlowRunObject], None]:
        """
        Gets paged results of `FlowRun` flavoured `DataObject`
        instances.

        Does not return count of results. [BUG]
        """

        if sort_by is not None:
            self.__raise_sort_by_error()

        self.__validate_object_type(object_type)

        size = page_size if page_size else self.get_page_size()

        filter_kwargs = self.__get_filter_kwargs(object_filters)

        return asyncio.run(
            self.__async_get_list_page(page_number, size, filter_kwargs)
        )

    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None
    ) -> Iterable[FlowRunObject]:

        page = 1

        while True:
            (results_page, _) = self.get_list_page(
                object_type,
                page,
                page_size=self.get_page_size(),
                object_filters=object_filters
            )

            if not results_page:
                return

            yield from results_page
            page += 1

    def insert(
        self,
        object_type: str,
        objects: Iterable[FlowRunObject]
    ) -> Iterable[FlowRunObject]:

        self.__validate_object_type(object_type)

        return asyncio.run(
            self.__async_insert(objects)
        )

    def insert_flow_run_iterable(
        self,
        objects: Iterable[FlowRunObject]
    ) -> Iterable[FlowRunObject]:
        """Inserts `FlowRunObject` instances into the prefect work queue"""

        return self.insert('flow_run', objects)

    def insert_flow_run(self, obj: FlowRunObject) -> FlowRunObject:
        """
        Inserts a single `FlowRunObject` instance into the prefect work queue
        """

        iter_result = self.insert_flow_run_iterable([obj])
        return list(iter_result)[0]

    async def get_names(
        self,
        deployment_id: Union[str, UUID]
    ) -> tuple[str, str]:
        """
        Given a `deployment_id`, returns the names
        of the corresponding flow and deployment.
        """

        async with self.__client_factory() as client:
            deployment = await client.read_deployment(
                deployment_id
            )
            flow_id = deployment.flow_id
            flow = await client.read_flow(flow_id)

        return flow.name, deployment.name

    async def get_deployment_id(
        self,
        flow_name: str,
        deployment_name: str
    ) -> Union[str, UUID]:
        """
        Given a `flow_name` and `deployment_name`, returns
        the UUID of the latter deployment.
        """

        name = f'{flow_name}/{deployment_name}'

        async with self.__client_factory() as client:
            deployment = await client.read_deployment_by_name(name)
            return deployment.id

    def __validate_object_type(self, object_type: str) -> None:
        if object_type != 'flow_run':
            self.__raise_bad_type_error()

    async def __async_get_by_id(
        self,
        object_ids: Iterable[str]
    ) -> Iterable[Optional[FlowRunObject]]:

        async with self.__client_factory() as client:
            fetched = [
                await self.__read_flow_run(client, id_)
                for id_ in object_ids
            ]

        return await self.__prc_factory().async_convert_iterable(fetched)

    async def __read_flow_run(
        self,
        client: PrefectClient,
        flow_run_id: Union[str, UUID]
    ) -> Optional[FlowRun]:

        try:
            return await client.read_flow_run(flow_run_id)
        except ObjectNotFound:
            return None

    async def __async_get_list_page(
        self,
        page_number: int,
        page_size: int,
        filter_kwargs: dict[str, Any]
    ) -> tuple[Iterable[FlowRunObject], None]:

        flow_runs = await self.__read_flow_runs(page_number, page_size, filter_kwargs)

        return await self.__prc_factory().async_convert_iterable(flow_runs), None

    async def __read_flow_runs(
        self,
        page_number: int,
        page_size: int,
        filter_kwargs: dict[str, Any]
    ) -> Iterable[FlowRun]:

        async with self.__client_factory() as client:
            return await client.read_flow_runs(
                sort=FlowRunSort.ID_DESC,
                limit=page_size,
                offset=(page_number - 1) * page_size,
                **filter_kwargs
            )

    def __get_filter_kwargs(
        self,
        object_filters: Optional[DataSourceFilter]
    ) -> dict[str, Any]:

        if object_filters is None:
            return {}

        return self.__filter_factory().to_kwargs(object_filters)

    async def __insert_kwargs_iterable(
        self,
        kwargs_iterable: Iterable[dict[str, Any]]
    ) -> Iterable[FlowRun]:

        async with self.__client_factory() as client:
            return [
                await self.__insert_kwargs(client, kwargs)
                for kwargs in kwargs_iterable
            ]

    async def __insert_kwargs(
        self,
        client: PrefectClient,
        kwargs: dict[str, Any]
    ) -> FlowRun:

        return await client.create_flow_run_from_deployment(
            **kwargs
        )

    async def __async_insert(
        self,
        objects: Iterable[FlowRunObject]
    ) -> Iterable[FlowRunObject]:

        kwargs_iter = await self.__oc_factory().async_convert_iterable(
            objects
        )
        flow_runs = await self.__insert_kwargs_iterable(kwargs_iter)
        return await self.__prc_factory().async_convert_iterable(
            flow_runs
        )

    def __raise_sort_by_error(self) -> None:
        raise DataSourceError(
            title='Sort Unsupported',
            detail='Sorting is unsupported on `PrefectDataSource`',
            status_code=400
        )

    def __raise_bad_type_error(self) -> None:
        raise DataSourceError(
            title='Bad Object Type',
            detail='Only "flow_run" is supported on `PrefectDataSource`.',
            status_code=400
        )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from functools import cache
from typing import Callable, Iterable, Optional

from .client import BoldApiClient
from .converter import (
    BoldApiConverter
)
from ..core import DataObject, DataSource, DataSourceError
from ..core.operator import (
    DetailGetter
)


ClientFactory = Callable[[], BoldApiClient]
BoldConverterFactory = Callable[[], BoldApiConverter]


class BoldDataSource(
    DataSource,

    # the supported operators
    DetailGetter
):
    """
    A `DataSource` that connects to a remote BOLD API.

    Developers should likely use `create_bold_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        bold_converter_factory: BoldConverterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__lc_factory = bold_converter_factory
        super().__init__({})

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'sample': {
                'processid': 'str',
                'record_id': 'str',
                'insdc_acs': 'str',
                'sampleid': 'str',
                'specimenid': 'int',
                'taxid': 'int',
                'short_note': 'str',
                'identification_method': 'str',
                'museumid': 'str',
                'fieldid': 'str',
                'collection_code': 'str',
                'processid_minted_date': 'datetime',
                'inst': 'str',
                'funding_src': 'str',
                'sex': 'str',
                'life_stage': 'str',
                'reproduction': 'str',
                'habitat': 'str',
                'collectors': 'str',
                'site_code': 'str',
                'specimen_linkout': 'str',
                'collection_event_id': 'str',
                'sampling_protocol': 'str',
                'tissue_type': 'str',
                'collection_date_start': 'datetime',
                'collection_time': 'str',
                'associated_taxa': 'str',
                'associated_specimens': 'str',
                'voucher_type': 'str',
                'notes': 'str',
                'taxonomy_notes': 'str',
                'collection_notes': 'str',
                'geoid': 'int',
                'marker_code': 'str',
                'kingdom': 'str',
                'phylum': 'str',
                'class': 'str',
                'order': 'str',
                'family': 'str',
                'subfamily': 'str',
                'tribe': 'str',
                'genus': 'str',
                'species': 'str',
                'subspecies': 'str',
                'taxon_name': 'str',
                'taxon_rank': 'str',
                'species_reference': 'str',
                'identified_by': 'str',
                'sequence_run_site': 'str',
                'nuc': 'str',
                'nuc_basecount': 'int',
                'sequence_upload_date': 'datetime',
                'bin_uri': 'str',
                'bin_created_date': 'datetime',
                'elev': 'int',
                'depth': 'int',
                'coord': 'List[int]',
                'coord_source': 'str',
                'coord_accuracy': 'str',
                'elev_accuracy': 'str',
                'depth_accuracy': 'str',
                'region': 'str',
                'sector': 'str',
                'site': 'str',
                'country_iso': 'str',
                'country/ocean': 'str',
                'province/state': 'str',
                'bold_recordset_code_arr': 'List[str]',
                'collection_date_end': 'datetime'
            },
            'bin': {
                'taxonomy': 'Dict[str, Dict[str, int]]'
            }
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        client = self.__client_factory()
        bold_response = client.get_detail(object_type, object_ids)
        bold_converter = self.__lc_factory()

        converted_objects, _ = bold_converter.convert_list(bold_response) \
            if bold_response is not None else ([], 0)
        yield from self.sort_by_id(converted_objects, object_ids)



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Optional

import flask

from ...core import DataSourceError


class NotAuthenticatedError(DataSourceError):
    """
    Raised when attempting to access `AuthContext`
    properties without a successful authentication event.
    """

    def __init__(self) -> None:
        """Initialize a NotAuthenticatedError with appropriate error details."""
        detail = 'No user has authenticated for this request'

        super().__init__(title='Unauthorized', detail=detail, status_code=401)


class AuthContext:
    """
    The auth context for a specific request, lasting
    only for its duration.

    Used principally to store the ID for the user making
    a request, if authenticated, but can also be used for
    other things.
    """

    def __init__(self) -> None:
        """Initialize an empty AuthContext with no authenticated user."""
        self.__user_id: Optional[str] = None
        self.__roles: list[str] = []
        self.__memberships: list[str] = []

    @property
    def authenticated(self) -> bool:
        """
        `True` if the user has authenticated successfully,
        perhaps by using a token.
        """

        return self.__user_id is not None

    @property
    def user_id(self) -> str:
        """A `str` that uniquely identifies a user."""

        self.__assert_authenticated()

        return self.__user_id

    @user_id.setter
    def user_id(self, val: str) -> None:
        """Set the user ID for this auth context."""
        self.__user_id = val

    @property
    def roles(self) -> list[str]:
        """
        A `list[str]` of names of roles assigned to
        the authenticated user.
        """

        self.__assert_authenticated()

        return self.__roles

    @roles.setter
    def roles(self, val: list[str]) -> None:
        """Set the roles for this auth context."""
        self.__roles = val

    def __assert_authenticated(self) -> None:
        """
        Assert that the user is authenticated.

        Raises:
            NotAuthenticatedError: If the user is not authenticated.
        """
        if not self.authenticated:
            raise NotAuthenticatedError()

    @property
    def memberships(self) -> list[str]:
        """
        A `list[str]` of memberships assigned to this authenticated user.
        """
        self.__assert_authenticated()

        return self.__memberships

    @memberships.setter
    def memberships(self, val: list[str]) -> None:
        """Set the memberships for this auth context."""
        self.__memberships = val


CtxGetter = Callable[[], AuthContext]
"""
A callable that fetches the global `AuthContext` instance
"""


def default_ctx_getter() -> AuthContext:
    """
    Get the default AuthContext instance for the current Flask request.

    Returns:
        AuthContext: The auth context stored in Flask's global context,
                    or a new instance if none exists.
    """
    return flask.g.setdefault('auth_context', default=AuthContext())



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from ..core import (
    core_data_object
)
from ..sql import (
    SqlDataSource,
    create_sql_datasource
)


# Create a SQL datasource. This is also expecting the models to be passed in
# so would most likely be used within an app
def sql(models, db_uri, behind_api, **kwargs) -> SqlDataSource:
    sql_ds = create_sql_datasource(
        models=models,
        db_uri=db_uri,
        behind_api=True,
        **kwargs
    )
    core_data_object(sql_ds)
    return sql_ds



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from abc import ABC, abstractmethod
from typing import Any, Optional, Union

from ...api_client.exception import BaseRuntimeException


Data = Union[
    dict[str, Any],
    list[dict[str, Any]]
]
Errors = list[dict[str, Any]]


class RequestBody(ABC):
    """A useful wrapper around a request body"""

    @property
    @abstractmethod
    def data(self) -> Optional[Data]:
        """Under the 'data' key"""

    @property
    @abstractmethod
    def meta(self) -> Optional[dict[str, Any]]:
        """Under the 'meta' key"""

    @property
    @abstractmethod
    def errors(self) -> Optional[Errors]:
        """Under the 'errors' key"""


class JsonApiRequestBody(RequestBody):
    """Implements RequestBody for a JSON:API body"""

    def __init__(self, json_dict: dict[str, Any]) -> None:
        self.__json = json_dict

    @property
    def data(self) -> Optional[Data]:
        if self.errors is not None:
            raise BaseRuntimeException(
                errors=[{
                    'title': 'Error Response Received',
                    'detail': json.dumps(self.errors)
                }]
            )
        return self.__json.get('data', {})

    @property
    def meta(self) -> Optional[dict[str, Any]]:
        return self.__json.get('meta')

    @property
    def errors(self) -> Optional[Errors]:
        return self.__json.get('errors')



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .client import DummyClient
from .converter import (
    DummyConverter
)
from .dummy_datasource import (
    DummyConverterFactory,
    DummyDataSource
)
from .parser import DefaultParser
from ..core import DataSource


class _DummyDSDict(Mapping):
    def __init__(self, api_ds: DummyDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> DummyDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manges the instantation of:

    - `DummyConverter`
    """

    def __init__(self) -> None:
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    def dummy_converter_factory(self) -> DummyConverterFactory:
        """
        Returns an instantiated `DummyConverter`.
        """

        parser = DefaultParser(self.__ds_dict)
        return DummyConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _DummyDSDict(self.data_source)


def _get_client_factory() -> Callable[[], DummyClient]:
    """
    A resonable default for creating
    a `DummyClient` instance
    """

    return lambda: DummyClient()


def create_dummy_datasource() -> DummyDataSource:
    """
    Instantiates `DummyDataSource`
    """

    client_factory = _get_client_factory()
    manager = _ConverterFactory()

    dummy_ds = DummyDataSource(
        client_factory,
        manager.dummy_converter_factory
    )

    manager.data_source = dummy_ds

    return dummy_ds



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from typing import List

from benchling_api_client.models.naming_strategy import NamingStrategy

from benchling_sdk.auth.api_key_auth import ApiKeyAuth
from benchling_sdk.benchling import Benchling
from benchling_sdk.errors import BenchlingError
from benchling_sdk.helpers.serialization_helpers import fields
from benchling_sdk.models import (
    CustomEntityBulkCreate,
    CustomEntityBulkUpdate
)

from .entities import convert_sts_entity_to_eln_entity_fields
from ..core import DataSource


class ElnDataSource(DataSource):

    def __init__(self, config):
        # url, api_key, registry_id, project_id, entities
        super().__init__(config, expected=['url', 'api_key', 'registry_id',
                                           'project_id', 'entities'])

        self.benchling_interface = self._get_benchling_interface(self.url, self.api_key)

    def _get_benchling_interface(self, url, api_key):
        return (Benchling(url=url, auth_method=ApiKeyAuth(api_key)))

    def __generate_response(self, task, entities, id_field):
        response = [{'id': entity[id_field], 'status': 'PASSED'} for entity in entities]
        try:
            if task.status == 'FAILED':
                for error in task.errors.additional_properties:
                    response[error['index']]['status'] = 'FAILED'
                    response[error['index']]['message'] = error['message']
                return response
            else:
                return response
        except BenchlingError as error:
            raise Exception(400, error.json['error']['message'])

    def register(self, entities, mapping_name):
        mapping = self.entities[mapping_name]
        schema_id = mapping['schema_id']
        name_field = mapping['name_field']
        request = []
        for entity in entities:
            entity_fields = convert_sts_entity_to_eln_entity_fields(entity, mapping)
            print(json.dumps(entity_fields, indent=4))
            name = entity[name_field]
            custom_fields = {}
            create_sample = CustomEntityBulkCreate(
                naming_strategy=NamingStrategy.REPLACE_NAMES_FROM_PARTS,
                schema_id=schema_id,
                name=name,
                fields=fields(entity_fields),
                folder_id=self.project_id,
                registry_id=self.registry_id,
                custom_fields=fields(custom_fields))
            request.append(create_sample)
        try:
            response = self.benchling_interface.custom_entities.bulk_create(request)
            task = self.benchling_interface.tasks.wait_for_task(
                response.task_id, interval_wait_seconds=3)
            print(f'{response.task_id}')
        except BenchlingError as error:
            raise Exception(400, error.json['error']['message'])
        response = [{'name': entity[name_field], 'status': 'PASSED'} for entity in entities]
        try:
            if task.status == 'FAILED':
                for error in task.errors.additional_properties:
                    response[error['index']]['status'] = 'FAILED'
                    response[error['index']]['message'] = error['message']
                return response
            else:
                return response
        except BenchlingError as error:
            raise Exception(400, error.json['error']['message'])

    def update(self, entities, mapping_name, **kwargs):
        mapping = self.entities[mapping_name]
        id_field = mapping['id_field']
        request = []
        for entity in entities:
            entity_fields = convert_sts_entity_to_eln_entity_fields(entity, mapping)
            custom_fields = {}
            update_sample = CustomEntityBulkUpdate(
                id=entity[id_field],
                fields=fields(entity_fields),
                custom_fields=fields(custom_fields))
            request.append(update_sample)
        try:
            response = self.benchling_interface.custom_entities.bulk_update(request)
            task = self.benchling_interface.tasks.wait_for_task(
                response.task_id, interval_wait_seconds=3)
        except BenchlingError as error:
            raise Exception(400, error.json['error']['message'])
        response = [{'id': entity[id_field], 'status': 'PASSED'} for entity in entities]
        try:
            if task.status == 'FAILED':
                response = [{'id': entity[id_field],
                             'status': 'FAILED',
                             'message': task.message} for entity in entities]
            return response
        except BenchlingError as error:
            raise Exception(400, error.json['error']['message'])

    @property
    def attribute_types(self):
        raise NotImplementedError()

    @property
    def supported_types(self) -> List:
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

"""
Composite authentication inspector module for managing authorization hooks and permissions.

This module provides a flexible authentication system that allows composition of multiple
authorization strategies and hooks for different object types and methods.
"""

import urllib
from collections import ChainMap
from itertools import chain
from typing import Optional

from flask import Blueprint, request

from .auth import AuthInspector
from .auth.error import AuthError
from .controller import Controller
from .misc import (
    ActionParameters,
    AggregationBody,
    AggregationParameters,
    GroupStatsParameters,
    JsonApiRequestBody,
    ListGetParameters,
    RelataionshipHopsParser,
    StatsParameters,
)
from ..api_client.exception import BaseRuntimeException
from ..api_client.parser import DefaultParser
from ..api_client.view import DefaultView
from ..core import DataSource, DataSourceError, OperableDataSource
from ..core.data_source_dict import DataSourceDict
from ..core.operator import Relational
from ..core.operator.operator_config import DefaultOperatorConfig, OperatorConfig
from ..core.requested_fields import ReqFieldsTree


class DataBlueprint(Blueprint):
    """
    A Flask Blueprint for dynamically routing DataObject endpoints defined in DataSource instances.

    This blueprint provides the foundation for creating REST API endpoints that interact with
    various data sources, handling common operations like GET, POST, PATCH, and DELETE requests.

    Args:
        url_prefix (str): The URL prefix for all routes registered to this blueprint.
    """

    def __init__(self, url_prefix: str) -> None:
        """
        Initialize the DataBlueprint.

        Args:
            url_prefix (str): URL prefix for all routes in this blueprint.
        """
        super().__init__('data_source_handler', __name__, url_prefix=url_prefix)


class ConfigBlueprint(Blueprint):
    """
    A Flask Blueprint for configuration endpoints, nested under DataBlueprint.

    This blueprint stores and serves configuration information about the stored data as a whole,
    including relationships, attribute types, metadata, and operational configurations.

    Args:
        url_prefix (str): The URL prefix for configuration routes.
    """

    def __init__(self, url_prefix: str) -> None:
        """
        Initialize the ConfigBlueprint.

        Args:
            url_prefix (str): URL prefix for configuration routes.
        """
        super().__init__('data_source_config', __name__, url_prefix=url_prefix)


class CustomBlueprint(Blueprint):
    """
    A Flask Blueprint for custom endpoints that need access to DataSources.

    This blueprint provides a flexible way to add custom endpoints that can leverage
    the existing DataSource infrastructure while implementing specialized business logic.

    Args:
        url_prefix (str): The URL prefix for custom routes.
        name (str, optional): The name for this blueprint. Defaults to __name__.
    """

    def __init__(self, url_prefix: str, name: str = __name__) -> None:
        """
        Initialize the CustomBlueprint.

        Args:
            url_prefix (str): URL prefix for custom routes.
            name (str, optional): Blueprint name. Defaults to __name__.
        """
        super().__init__(name, __name__, url_prefix=url_prefix)


def _config_blueprint(
    url_prefix: str,
    data_sources: tuple[OperableDataSource],
    operator_config: OperatorConfig,
) -> ConfigBlueprint:
    """
    Create and configure a ConfigBlueprint instance with configuration endpoints.

    This function sets up various configuration endpoints that provide metadata about
    the data sources, including relationships, attribute types, operations, and modes.

    Args:
        url_prefix (str): The URL prefix on which to serve the configuration endpoints.
        data_sources (tuple[OperableDataSource]): A tuple of DataSource instances behind the API.
        operator_config (OperatorConfig): Configuration for supported operations.

    Returns:
        ConfigBlueprint: A configured blueprint instance with all configuration routes.

    Routes created:
        - GET /relationships: Returns relationship configurations for relational data sources
        - GET /attribute_types: Returns combined attribute types from all data sources
        - GET /attribute_metadata: Returns combined attribute metadata from all data sources
        - GET /operations: Returns available operations configuration
        - GET /write_mode: Returns write mode configurations for data sources
        - GET /return_mode: Returns return mode configurations for data sources
    """
    config_handler = ConfigBlueprint(url_prefix)

    @config_handler.route('/relationships', methods=['GET'])
    def get_relationships():
        """Get relationship configurations from all relational data sources."""
        relationship_configs = chain(
            *[d.relationship_config.items() for d in data_sources if isinstance(d, Relational)]
        )
        return {t: d.to_dict() for t, d in relationship_configs if not d.empty}

    @config_handler.route('/attribute_types', methods=['GET'])
    def get_attribute_types():
        """Get combined attribute types from all data sources."""
        types_list = [d.attribute_types for d in data_sources]
        chain_map = ChainMap(*types_list)
        return dict(chain_map)

    @config_handler.route('/attribute_metadata', methods=['GET'])
    def get_attribute_metadata():
        """Get combined attribute metadata from all data sources."""
        types_list = [d.attribute_metadata for d in data_sources]
        chain_map = ChainMap(*types_list)
        return dict(chain_map)

    @config_handler.route('/operations', methods=['GET'])
    def get_operations():
        """Get available operations configuration."""
        return operator_config.to_dict()

    @config_handler.get('/write_mode')
    def get_relation_write_mode():
        """Get write mode configurations for all data sources."""
        modes_list = [d.write_mode for d in data_sources]
        chain_map = ChainMap(*modes_list)
        return dict(chain_map)

    @config_handler.get('/return_mode')
    def get_return_mode():
        """Get return mode configurations for all data sources."""
        modes_list = [d.return_mode for d in data_sources]
        chain_map = ChainMap(*modes_list)
        return dict(chain_map)

    return config_handler


def _core_blueprint(
    data_source_dict: dict[str, DataSource],
    url_prefix: str,
    auth_inspector: Optional[AuthInspector] = None,
    include_all_to_ones: bool = True,
    flow_ds: Optional[OperableDataSource] = None,
    action_ds: Optional[OperableDataSource] = None,
) -> DataBlueprint:
    """
    Create the core blueprint responsible for managing DataSource endpoints.

    This function creates a comprehensive set of REST API endpoints for interacting
    with DataSource instances, including CRUD operations, relationships, aggregations,
    and specialised operations like cursor-based pagination.

    Args:
        data_source_dict (dict[str, DataSource]): Dictionary mapping object
            types to DataSource instances.
        url_prefix (str): URL prefix for all data endpoints.
        auth_inspector (Optional[AuthInspector], optional): Authentication
            inspector for request authorisation.
        include_all_to_ones (bool): Whether to fetch or store all to-one related objects
            when fetching or serialising DataObjects.
        flow_ds (Optional[DataSource], optional): DataSource for flow operations.
    Returns:
        DataBlueprint: A configured blueprint with all data endpoints and error handlers.

    Routes created:
        - GET /<object_type>/<object_id>: Get individual object details
        - GET /<object_type>: Get paginated list of objects
        - GET /<object_type>:count: Get count of objects matching filters
        - GET /<object_type>:stats: Get statistical information about objects
        - GET /<object_type>:group-stats: Get grouped statistical information
        - DELETE /<object_type>/<object_id>: Delete individual object
        - PATCH /<object_type>: Update multiple objects
        - POST /<object_type>:insert: Insert new objects
        - POST /<object_type>:upsert: Insert or update objects
        - POST /<object_type>:aggregations: Get aggregated data
        - POST /<object_type>:cursor: Get cursor-based pagination
        - GET /<object_type>:to-one/<object_id>/<hops_suffix>: Navigate to-one relationships
        - GET /<object_type>:to-many/<object_id>/<relationship_name>: Get to-many relationships
    """
    data_handler = DataBlueprint(url_prefix=url_prefix)

    def __new_controller(
        object_type: str,
        requested_fields: list[str] | None = None,
    ) -> Controller:
        """
        Create a new Controller instance for handling requests to a specific object type.

        Args:
            object_type (str): The type of object this controller will handle.
            requested_fields (list[str] | None, optional): Specific fields to include in responses.

        Returns:
            Controller: Configured controller instance.
        """
        hop_limit = None if requested_fields else 1

        data_source = data_source_dict[object_type]

        # Build a ReqFieldsTree template for the request
        req_fields_tree = ReqFieldsTree(
            object_type,
            data_source,
            requested_fields,
            include_all_to_ones=include_all_to_ones,
        )

        view = DefaultView(
            requested_tree=req_fields_tree,
            prefix=url_prefix,
            hop_limit=hop_limit,
        )
        return Controller(data_source, view, req_fields_tree, auth_inspector=auth_inspector)

    def __new_parser(
        object_type: str,
    ):
        data_source = data_source_dict[object_type]
        # Build a ReqFieldsTree template for the request
        req_fields_tree = ReqFieldsTree(
            object_type,
            data_source,
            include_all_to_ones=include_all_to_ones,
        )
        return DefaultParser(data_source_dict, requested_tree=req_fields_tree)

    @data_handler.route('/<object_type>/<path:object_id>', methods=['GET'])  # Allow slashes
    def get_detail(*, object_type: str, object_id: str):
        """Get details of a specific object by ID."""
        request_args = ListGetParameters(request.args)
        controller = __new_controller(
            object_type,
            requested_fields=request_args.requested_fields,
        )
        object_id_unencoded = urllib.parse.unquote(object_id)
        return controller.get_detail(object_type, object_id_unencoded)

    @data_handler.route('/<object_type>', methods=['GET', 'POST'])
    def get_list(*, object_type: str):
        """Get a paginated list of objects of the specified type."""
        if request.method == 'POST':
            request_args = ListGetParameters(request.json | request.args)
        else:
            request_args = ListGetParameters(request.args)
        controller = __new_controller(
            object_type,
            requested_fields=request_args.requested_fields,
        )
        return controller.get_list(object_type, request_args)

    @data_handler.route('/<object_type>:count', methods=['GET', 'POST'])
    def get_count(*, object_type: str):
        """Get the count of objects matching the specified filters."""
        if request.method == 'POST':
            request_args = ListGetParameters(request.json | request.args)
        else:
            request_args = ListGetParameters(request.args)
        controller = __new_controller(object_type)
        return controller.get_count(object_type, request_args)

    @data_handler.route('/<object_type>:stats', methods=['GET', 'POST'])
    def get_stats(*, object_type: str):
        """Get statistical information about objects of the specified type."""
        if request.method == 'POST':
            request_args = StatsParameters(request.json | request.args)
        else:
            request_args = StatsParameters(request.args)
        controller = __new_controller(object_type)
        return controller.get_stats(object_type, request_args)

    @data_handler.get('/<object_type>:group-stats')
    def get_group_stats(*, object_type: str):
        """Get grouped statistical information about objects."""
        controller = __new_controller(object_type)
        request_args = GroupStatsParameters(request.args)
        return controller.get_group_stats(object_type, request_args)

    @data_handler.route('/<object_type>/<path:object_id>', methods=['DELETE'])
    def delete_detail(*, object_type: str, object_id: str):
        """Delete a specific object by ID."""
        controller = __new_controller(object_type)
        object_id_unencoded = urllib.parse.unquote(object_id)
        return controller.delete_detail(object_type, object_id_unencoded)

    @data_handler.route('/<object_type>', methods=['PATCH'])
    def patch_list(*, object_type: str):
        """Update multiple objects of the specified type."""
        controller = __new_controller(object_type)
        request_body = JsonApiRequestBody(request.json)
        return controller.patch_list(object_type, request_body.data)

    @data_handler.post('/<object_type>:insert')
    def post_inserts(*, object_type: str):
        """Insert new objects of the specified type."""
        controller = __new_controller(object_type)
        parser = __new_parser(object_type)
        objects = parser.parse_json_doc(request.json)
        return controller.post_inserts(object_type, objects)

    @data_handler.route('/<object_type>:upsert', methods=['POST'])
    def post_upserts(*, object_type: str):
        """Insert or update objects of the specified type."""
        request_args = ListGetParameters(request.args)
        controller = __new_controller(object_type)
        parser = __new_parser(object_type)
        objects = parser.parse_json_doc(request.json)
        return controller.post_upserts(
            object_type,
            objects,
            merge_collections=request_args.merge_collections,
        )

    @data_handler.route('/<object_type>:aggregations', methods=['POST'])
    def get_aggregations(*, object_type: str):
        """Get aggregated data for objects of the specified type."""
        controller = __new_controller(object_type)
        request_args = AggregationParameters(request.args | request.json)
        body = AggregationBody(request.json)
        return controller.post_aggregations(object_type, request_args, body)

    @data_handler.post('/<object_type>:cursor')
    def get_cursor_page(*, object_type: str):
        """Get a page of results using cursor-based pagination."""
        request_args = ListGetParameters(request.args | request.json)
        controller = __new_controller(
            object_type,
            requested_fields=request_args.requested_fields,
        )
        search_after = request.json.get('search_after')
        page = controller.get_cursor_page(object_type, request_args, search_after)
        return page

    @data_handler.post('/<object_type>:action')
    def post_action(*, object_type: str):
        """Perform an action on objects of the specified type."""
        request_args = ActionParameters(request.json | request.args)
        controller = __new_controller(object_type)
        return controller.perform_action(object_type, request_args, action_ds, flow_ds)

    @data_handler.route('/<object_type>:to-one/<object_id>/<path:hops_suffix>', methods=['GET'])
    def get_to_one_relation(*, object_type: str, object_id: str, hops_suffix: str):
        """Navigate through to-one relationships following the specified path."""
        controller = __new_controller(object_type)
        source = data_source_dict[object_type].data_object_factory(object_type, object_id)
        hops = RelataionshipHopsParser(hops_suffix).relationship_hops
        return controller.get_recursive_relation(source, hops)

    @data_handler.route('/<object_type>:to-many/<object_id>/<relationship_name>', methods=['GET'])
    def get_to_many_relations(*, object_type: str, object_id: str, relationship_name: str):
        """Get objects related through a to-many relationship."""
        controller = __new_controller(object_type)
        source = data_source_dict[object_type].data_object_factory(object_type, object_id)
        params = ListGetParameters(request.args)
        return controller.get_many_relations_page(source, relationship_name, params)

    @data_handler.app_errorhandler(BaseRuntimeException)
    def handle_runtime_error(error: BaseRuntimeException):
        """Handle base runtime exceptions and format error responses."""
        return {'errors': error.errors}, error.status_code

    @data_handler.app_errorhandler(DataSourceError)
    def handle_datasource_error(error: DataSourceError):
        """Handle data source errors and format error responses."""
        return {'errors': [{'title': error.title, 'detail': error.detail}]}, error.status_code

    @data_handler.app_errorhandler(AuthError)
    def handle_auth_error(error: AuthError):
        """Handle authentication/authorization errors and format error responses."""
        return {'errors': error.errors}, error.status_code

    return data_handler


def data_blueprint(
    *data_sources: DataSource,
    flow_ds: Optional[DataSource] = None,
    action_ds: Optional[DataSource] = None,
    url_prefix: str = '/data',
    config_prefix: str = '/_config',
    auth_inspector: Optional[AuthInspector] = None,
    include_all_to_ones: bool = True,
) -> DataBlueprint:
    """
    Create a complete data blueprint with both core and configuration endpoints.

    This is the main factory function for creating a full-featured data API blueprint
    that includes both data manipulation endpoints and configuration/metadata endpoints.

    Args:
        *data_sources (DataSource): Variable number of DataSource instances to expose.
        url_prefix (str, optional): URL prefix for data endpoints. Defaults to '/data'.
        config_prefix (str, optional): URL prefix for configuration endpoints.
            Defaults to '/_config'.
        auth_inspector (Optional[AuthInspector], optional): Authentication inspector.
            Defaults to None.

    Returns:
        DataBlueprint: A complete blueprint with both data and configuration endpoints.

    Example:
        ```python
        # Create blueprint with multiple data sources
        blueprint = data_blueprint(
            user_datasource,
            project_datasource,
            url_prefix='/api/v1/data',
            auth_inspector=my_auth_inspector
        )

        # Register with Flask app
        app.register_blueprint(blueprint)
        ```
    """
    config_bp = _config_blueprint(
        config_prefix, data_sources, DefaultOperatorConfig(*data_sources)
    )
    core_bp = _core_blueprint(
        DataSourceDict(*data_sources),
        url_prefix,
        auth_inspector=auth_inspector,
        include_all_to_ones=include_all_to_ones,
        flow_ds=flow_ds,
        action_ds=action_ds,
    )
    core_bp.register_blueprint(config_bp)

    return core_bp


def custom_blueprint(url_prefix: str = '/custom', name: str = 'custom') -> DataBlueprint:
    """
    Create a Flask Blueprint instance for adding custom endpoints.

    This function provides a basic blueprint that can be extended with custom routes
    while still having access to the DataSource infrastructure if needed.

    Args:
        url_prefix (str, optional): URL prefix for custom endpoints. Defaults to '/custom'.
        name (str, optional): Name for the blueprint. Defaults to 'custom'.

    Returns:
        DataBlueprint: A blueprint instance ready for custom endpoint registration.

    Example:
        ```python
        # Create custom blueprint
        custom_bp = custom_blueprint('/api/v1/custom', 'my_custom_endpoints')

        # Add custom routes
        @custom_bp.route('/health', methods=['GET'])
        def health_check():
            return {'status': 'healthy'}

        # Register with Flask app
        app.register_blueprint(custom_bp)
        ```
    """
    custom_handler = CustomBlueprint(name=name, url_prefix=url_prefix)

    return custom_handler



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from sqlalchemy import cast
from sqlalchemy.types import BigInteger, Float, String, TypeDecorator


class CastToBigIntegerType(TypeDecorator):
    impl = String
    python_type = int

    def column_expression(self, col):
        return cast(col, BigInteger)

    def process_bind_param(self, value, dialect):
        return str(value)


class CastToFloatType(TypeDecorator):
    impl = String
    python_type = float

    def column_expression(self, col):
        return cast(col, Float)

    def process_bind_param(self, value, dialect):
        return str(value) if value is not None else None



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class TreeofsexUploadToTreeofsexwhSpeciesConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        ret = self._data_object_factory(
            'species',
            data_object.species,
            attributes={
                data_object.key: [
                    {
                        'value': data_object.value,
                        'source': data_object.reference,
                    }
                ]
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class CombineFieldsConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field1: str
        field2: str
        dest_field: str
        lowercase_field1: bool

    __slots__ = ('__config',)
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        Concatenates the values of two fields and stores the result in a new field.
        The first field's value may be lowercased if specified in the configuration.
        The destination field name is given by the configuration.
        """

        val1 = data_object.get_field_by_name(self.__config.field1)
        val2 = data_object.get_field_by_name(self.__config.field2)
        attributes = dict(data_object.attributes)

        if val1 is not None and val2 is not None:
            part1 = str(val1).lower() if self.__config.lowercase_field1 else str(val1)
            attributes[self.__config.dest_field] = f'{part1}{val2}'

        yield self._data_object_factory(
            data_object.type,
            data_object.id,
            attributes=attributes,
        )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_s3_to_https(s3_path):
    if s3_path.startswith('s3://'):
        no_prefix = s3_path[len('s3://'):]
        parts = no_prefix.split('/', 1)
        if len(parts) == 2:
            bucket, path = parts
            return f'https://{bucket}.cog.sanger.ac.uk/{path}'
        else:
            # Only bucket, no path
            return f'https://{parts[0]}.cog.sanger.ac.uk/'
    return s3_path



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
from dataclasses import dataclass
from pathlib import Path

from jsonschema import Draft202012Validator
from jsonschema.exceptions import ValidationError

from tol.core import DataObject
from tol.core.validate import Validator

EXPECTED_DRAFT = 'https://json-schema.org/draft/2020-12/schema'


class JsonSchemaValidator(Validator):
    """
    Validate JSON content of a DataObject against a Draft 2020-12 schema.
    Also enforces that the schema declares Draft 2020-12 via the $schema field.
    """

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        schema_path: Path
        field: str = 'jsondata'
        is_error: bool = True
        detail: str = 'JSON schema validation failed'

    __slots__ = ['__config', '__validator']
    __config: Config
    __validator: Draft202012Validator

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        del kwargs
        super().__init__()
        self.__config = config

        with open(config.schema_path) as f:
            schema = json.load(f)

        actual_draft = schema.get('$schema')
        if actual_draft != EXPECTED_DRAFT:
            raise ValueError(
                f'Schema draft version mismatch. Expected "{EXPECTED_DRAFT}", got "{actual_draft}"'
            )

        Draft202012Validator.check_schema(schema)

        self.__validator = Draft202012Validator(schema)

    def _validate_data_object(
        self,
        obj: DataObject,
    ) -> None:

        data = getattr(obj, self.__config.field, None)

        if data is None:
            self.__add_result(
                obj=obj,
                field=self.__config.field,
                detail=f'Missing field "{self.__config.field}" in DataObject'
            )
            return

        error: ValidationError = next(self.__validator.iter_errors(data), None)
        if error is not None:
            self.__add_result(
                obj=obj,
                field=self.__config.field,
                detail=f'{self.__config.detail}: {error.message}'
            )

    def __add_result(
        self,
        obj: DataObject,
        field: str,
        detail: str
    ) -> None:
        if self.__config.is_error:
            self.add_error(
                object_id=obj.id,
                detail=detail,
                field=field
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=detail,
                field=field
            )



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from tol.core import DataObject, Validator
from tol.sources.goat import GoatDataSource, goat


class TaxonMatchesGoatValidator(Validator):
    """
    Validates a stream of `DataObject` instances, checking whether its Taxonomy information
    matches that in GoaT
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        species_field: str | None = None
        genus_field: str | None = None
        family_field: str | None = None
        superfamily_field: str | None = None
        phylum_field: str | None = None
        kingdom_field: str | None = None
        superkingdom_field: str | None = None
        domain_field: str | None = None

    __slots__ = ['__config', '__goat_datasource', '_cached_taxa']
    __config: Config
    __goat_datasource: GoatDataSource
    _cached_taxa: dict[str, DataObject]

    def __init__(self, config: Config, **kwargs) -> None:
        super().__init__()
        self.__config = config
        self.__goat_datasource = goat()
        self._cached_taxa = {}

    def _validate_data_object(self, obj: DataObject) -> None:
        taxon_id = obj.get_field_by_name('TAXON_ID')

        # Check whether we already have the information for this id in the cache.
        # If we don't, fetch it from GoaT and add it to the cache
        taxon: DataObject | None
        if taxon_id in self._cached_taxa:
            taxon = self._cached_taxa[taxon_id]
        else:
            taxon = self.__goat_datasource.get_one('taxon', taxon_id)

            # Add this taxon to cached taxa.
            # Error if GoaT has no taxon with this id
            if taxon is not None:
                self._cached_taxa[taxon_id] = taxon
            else:
                self.add_error(
                    object_id=obj.id,
                    detail=f'Invalid Taxon ID: {taxon_id}',
                    field='taxon_id'
                )

                # We can't validate a taxon that doesn't exist, so after this error move on
                # to the next DataObject
                return

        # Check that each associated taxonomy rank for this taxon matches those in GoaT
        taxonomic_ranks = ('species', 'genus', 'family', 'superfamily',
                           'phylum', 'kingdom', 'superkingdom', 'domain')
        for rank in taxonomic_ranks:
            # From the rank to check, get the name of its field in the data object we're validating
            # from the validator config. If this field name is `None`, then this taxonomic rank
            # isn't being checked (likely because the data object does not have this field)
            field_name: str | None = getattr(self.__config, f'{rank}_field')
            if field_name is None:
                continue

            # Fetch the values of these taxonomic ranks
            value_in_data_object = obj.get_field_by_name(field_name)
            value_in_goat = taxon.get_field_by_name(f'{rank}.scientific_name')

            # Ensure the value in the data object matches the one in GoaT
            if value_in_data_object != value_in_goat:
                self.add_warning(
                    object_id=obj.id,
                    detail=(f'Value for {field_name} ({value_in_data_object}) '
                            f'does not match the value in GoaT ({value_in_goat})'),
                    field=field_name,
                )



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from functools import cache
from typing import Callable, Iterable, Optional

from .client import LabwhereApiClient
from .converter import (
    LabwhereApiConverter
)
from ..core import DataObject, DataSource, DataSourceError
from ..core.operator import (
    DetailGetter,
    Relational
)
from ..core.relationship import RelationshipConfig


ClientFactory = Callable[[], LabwhereApiClient]
LabwhereConverterFactory = Callable[[], LabwhereApiConverter]


class LabwhereDataSource(
    DataSource,

    # the supported operators
    DetailGetter,
    Relational
):
    """
    A `DataSource` that connects to a remote LabWhere API.

    Developers should likely use `create_labwhere_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        labwhere_converter_factory: LabwhereConverterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__lc_factory = labwhere_converter_factory
        super().__init__({})

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            'location': {
                'name': 'str',
                'parent': 'str',
                'container': 'bool',
                'status': 'str',
                'rows': 'int',
                'columns': 'int',
                'parentage': 'str',
                'location_type_id': 'int',
                'created_at': 'datetime',
                'updated_at': 'datetime'
            },
            'location_type': {
                'name': 'str',
                'created_at': 'datetime',
                'updated_at': 'datetime'
            }
        }

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        rc_location = RelationshipConfig()
        rc_location.to_one = {
            'parent_location': 'location',
            'location_type': 'location_type'
        }
        # rc_location.to_many = {'child_locations': 'location'}
        rc_location_type = RelationshipConfig()
        # rc_location_type.to_many = {'locations': 'location'}
        return {
            'location': rc_location,
            'location_type': rc_location_type
        }

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        **kwargs,
    ) -> Iterable[Optional[DataObject]]:
        if object_type not in self.supported_types:
            raise DataSourceError(f'{object_type} is not supported')

        client = self.__client_factory()
        labwhere_responses = (
            client.get_detail(object_type, id_)
            for id_ in object_ids
        )
        labwhere_converter = self.__lc_factory()
        return (
            labwhere_converter.convert(r)
            if r is not None else None
            for r in labwhere_responses
        )

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Optional[DataObject]:

        if source.type == 'location' and relationship_name == 'parent_location':
            if source.parent is not None:
                parentage_ids = source.parent.split('/')
                parent_id = parentage_ids[len(parentage_ids) - 1]
                if parent_id != 'Empty':
                    return self.get_one(
                        'location',
                        parent_id
                    )
        if source.type == 'location' and relationship_name == 'location_type':
            if source.location_type_id is not None:
                return self.get_one(
                    'location_type',
                    source.location_type_id
                )

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str
    ) -> Iterable[DataObject]:

        return []



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing

from flask import Blueprint

from ..api_base.auth import ForbiddenError
from ..api_base.misc import (
    CtxGetter,
    default_ctx_getter
)
from ..core import (
    DataObject,
    DataSourceError,
    DataSourceFilter
)

if typing.TYPE_CHECKING:
    from ..sql import SqlDataSource


TYPE_HIERARCHY = [
    'board',
    'view',
    'zone',
    'component',
]


def board_blueprint(
    board_ds: SqlDataSource,
    type_hierarchy: list[str] = TYPE_HIERARCHY,

    ctx_getter: CtxGetter = default_ctx_getter,
) -> Blueprint:
    """
    Provides a flask `Blueprint` for management of
    user-configurable dashboarding resources.
    """

    board_bp = Blueprint(
        'dashboards',
        __name__
    )

    smallest_type = type_hierarchy[-1]
    biggest_type = type_hierarchy[0]

    def __smaller_is_deletable(
        smaller_obj: DataObject,
        bigger_type: str,
        all_bigger_ids: list[str],
        joiner_type: str
    ) -> bool:
        """
        Reasons a smaller row can't be deleted:

        1. It doesn't belong to the user that
           is initiating the delete (managed by
           the calling function before this one).

        2. Other bigger rows point to this smaller
           (e.g. this `zone` is in another `view`).
        """

        f = DataSourceFilter(
            and_={
                f'{bigger_type}.id': {
                    'in_list': {
                        'value': all_bigger_ids,
                        'negate': True
                    }
                },
                f'{smaller_obj.type}.id': {
                    'eq': {
                        'value': smaller_obj.id
                    }
                },
            }
        )
        count = board_ds.get_count(
            joiner_type,
            object_filters=f
        )

        return count == 0

    def __get_deletable_smallers(
        smaller_type: str,
        joiner_type: str,
        bigger_type: str,
        all_bigger_ids: list[str],
        joins: list[DataObject],
        user_id: str
    ) -> list[DataObject]:
        """
        Given a bigger->smaller relation (e.g.
        `zone` rows in a `view`) and the join
        rows that define it (e.g. `zone_view`),
        this function gets the smaller rows
        (here `zone`) that can be deleted.
        """

        all_smaller_objs: list[DataObject] = [
            getattr(join, smaller_type)
            for join in joins
        ]

        return [
            obj for obj in all_smaller_objs
            if obj.user.id == user_id
            and __smaller_is_deletable(
                obj,
                bigger_type,
                all_bigger_ids,
                joiner_type
            )
        ]

    def __delete_recursive(
        bigger_type: str,
        bigger_objs: list[DataObject],
        user_id: str
    ) -> None:
        """
        Given a list of bigger, containing objects (e.g. `view`),
        deletes the contained objects (`zone`->`component`)
        recursively.

        Stops recursion within a branch in which its head can't
        be deleted.
        """

        all_bigger_ids = [obj.id for obj in bigger_objs]

        if bigger_type != smallest_type:
            all_deletable_smallers = []
            all_join_ids = []

            bigger_index = type_hierarchy.index(bigger_type)
            smaller_type = type_hierarchy[bigger_index + 1]
            joiner_type = f'{smaller_type}_{bigger_type}'

            for bigger_obj in bigger_objs:
                joins_filter = DataSourceFilter(
                    and_={
                        f'{bigger_obj.type}.id': {
                            'eq': {
                                'value': bigger_obj.id
                            }
                        }
                    }
                )
                joins = list(
                    board_ds.get_list(
                        joiner_type,
                        object_filters=joins_filter
                    )
                )

                deletable_smallers = __get_deletable_smallers(
                    smaller_type,
                    joiner_type,
                    bigger_type,
                    all_bigger_ids,
                    joins,
                    user_id
                )
                all_deletable_smallers.extend(deletable_smallers)

                join_ids = [j.id for j in joins]
                all_join_ids.extend(join_ids)

            # delete the joins first
            board_ds.delete(joiner_type, all_join_ids)

            __delete_recursive(smaller_type, all_deletable_smallers, user_id)

        board_ds.delete(bigger_type, all_bigger_ids)

    def __delete_above(
        object_type: str,
        object_id: str,
        user_id: str
    ) -> None:
        """
        Deletes the (sole) joining table entry pointing to the specified
        `object_type`, if it's not the biggest type (aka `board`).

        Fails if:
        - there is more than one joining entry (e.g. `zone_view` -> `zone`)
        - the (sole) joining entry does not belong to the authenticated user
        """

        if object_type == biggest_type:
            return

        object_index = type_hierarchy.index(object_type)
        above_type = type_hierarchy[object_index - 1]
        joiner_type = f'{object_type}_{above_type}'

        f = DataSourceFilter(
            and_={
                f'{object_type}.id': {
                    'eq': {
                        'value': object_id
                    }
                }
            }
        )

        above_count = board_ds.get_count(joiner_type, object_filters=f)
        if above_count == 0:
            return
        if above_count > 1:
            raise DataSourceError(
                'Deletion Error',
                f'More than one {above_type}s instances point '
                f'to this {object_type}.',
                400
            )

        (joiner_obj,) = list(
            board_ds.get_list(
                joiner_type,
                object_filters=f
            )
        )

        above_obj: DataObject = getattr(joiner_obj, above_type)
        if above_obj.user.id != user_id:
            raise DataSourceError(
                'Deletion Error',
                f'The linked {above_type} is not yours.',
                400
            )

        board_ds.delete(joiner_type, [joiner_obj.id])

    def delete(
        bigger_type: str,
        bigger_id: str,
        user_id: str
    ) -> None:
        """
        Given a bigger, containing object (e.g. `view`):

        - Deletes the sole join to an above object if one exists
          (here `board`). Raises a `DataSourceError` if either:
            1. the above object does not belong to the user
               calling this method.
            2. there is more than one above join (e.g. if this
               `zone` is in more than one `board`) regardless
               of user-ownership.
        - Recursively deletes all descendents (here
          `zone`->`component`), ending branching at any node
          that can't be deleted.
        - Deletes this bigger, containing object (here `view`).
        """

        bigger_obj = board_ds.get_one(bigger_type, bigger_id)

        if bigger_obj is None:
            raise DataSourceError(
                'Not Found',
                f'The given {bigger_type} was not found.',
                404
            )

        if bigger_obj.user.id != user_id:
            raise ForbiddenError()

        __delete_above(
            bigger_type,
            bigger_id,
            user_id
        )

        __delete_recursive(
            bigger_type,
            [bigger_obj],
            user_id
        )

    @board_bp.delete('/<string:object_type>/<string:object_id>')
    def __delete_endpoint(*, object_type: str, object_id: str):
        if object_type not in type_hierarchy:
            raise DataSourceError(
                'Unknown Type',
                'The given type is not recognised in the hierarchy',
                400
            )

        delete(
            object_type,
            object_id,
            ctx_getter().user_id
        )

        return {'deleted': True}, 200

    return board_bp



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from itertools import islice
from typing import Iterable, Optional

from ...core.datasource_error import DataSourceError

if typing.TYPE_CHECKING:
    from ..data_object import DataObject
    from ..relationship import RelationshipConfig
    from ..requested_fields import ReqFieldsTree
    from ..session import OperableSession


class Relational(ABC):
    """
    Augments a DataSource to support relationships between hosted
    DataObject types.
    """

    @property
    @abstractmethod
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        """
        The configuration of relationships (both to-one and to-many) between
        the types of DataObject instances managed by this DataSource instance.
        """

    @abstractmethod
    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Optional[DataObject]:
        """
        Gets the to-one relation DataObject, given a source DataObject and the
        name of the relationship within the config.
        """

    @abstractmethod
    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:
        """
        Gets the Iterable of to-many relation DataObject instances, given a source
        DataObject and the name of the relationship within the config.
        """

    def get_to_many_relations_page(
        self,
        source: DataObject,
        relationship_name: str,
        page: int,
        page_size: int,
        session: Optional[OperableSession] = None,
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:
        """
        Slices the `Relational().get_to_many_relations()` `Iterable` into pages.

        This can be overriden if a more efficient method is available.
        """

        iter_many = self.get_to_many_relations(source, relationship_name)
        start = (page - 1) * page_size
        stop = page * page_size

        return islice(iter_many, start, stop)

    def get_recursive_relation(
        self,
        source: DataObject,
        relationship_hops: list[str],
        requested_fields: list[str] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> DataObject:
        """
        Recursively get to-one relation `DataObject` instances, using the
        names defined in `relationship_hops`, returning the last one.
        """

        if not relationship_hops:
            return source

        first_hop = relationship_hops[0]
        target = self.get_to_one_relation(source, first_hop)

        if target is None:
            return None

        return self.get_recursive_relation(target, relationship_hops[1:])

    def validate_to_one_recurse(
        self,
        source_type: str,
        relationship_hops: list[str]
    ) -> None:
        """Raises a `DataSourceError` on a bad relationship name"""

        if not relationship_hops:
            return

        relationship_name = relationship_hops[0]
        self.validate_to_one_relationship(source_type, relationship_name)
        target_type = self.__get_relationship_target(
            source_type,
            relationship_name
        )
        self.validate_to_one_recurse(target_type, relationship_hops[1:])

    def validate_to_one_relationship(
        self,
        object_type: str,
        relationship_name: str
    ) -> None:
        """
        Validates that the specified `relationship_name` is defined on
        the given `object_type`.
        """

        to_one = self.relationship_config[object_type].to_one

        if relationship_name not in to_one:
            self.__raise_bad_to_one_relationsip(object_type, relationship_name)

    def __raise_bad_to_one_relationsip(
        self,
        object_type: str,
        relationship_name: str
    ) -> None:

        detail = (
            f'No to-one relationship "{relationship_name}" '
            f'exists on type {object_type}.'
        )
        raise DataSourceError(
            title='Bad Relationship Name',
            detail=detail,
            status_code=400
        )

    def __get_relationship_target(
        self,
        object_type: str,
        relationship_name: str
    ) -> str:

        to_one = self.relationship_config[object_type].to_one
        return to_one[relationship_name]



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class MlwhExtractionToElasticExtractionConverter(
        DataObjectToDataObjectOrUpdateConverter):

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        prefix = data_object.assay_type.lower()
        if prefix[0] == '_':
            prefix = prefix[1:]
        value = data_object.value
        try:
            value = float(value)
        except (ValueError, TypeError):
            pass  # Keep the original value if conversion fails
        ret = self._data_object_factory(
            'extraction',
            data_object.sample_id,
            attributes={
                'tissue_prep_fluidx_id': data_object.labware_barcode,
                f'{prefix}_value': value,
                f'{prefix}_units': data_object.units,
                f'{prefix}_recorded_at': data_object.recorded_at,
                'qc_status': data_object.qc_status,
                'qc_status_decision_by': data_object.qc_status_decision_by
            }
        )
        yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Any

import numpy as np

import pandas as pd

from .pandas_datasource import PandasDataSource


class CsvDataSource(
    PandasDataSource
):
    """
    Emits `DataObject` instances from a tabular
    spreadsheet (e.g. `.xlsx`).
    """

    def __init__(
        self,
        filepath: str | Path,
        *,
        object_type: str = 'sheet_row',
        type_mapping: dict[str, str] | None = None,
    ) -> None:

        super().__init__(
            filepath,
            sheet_name=None,
            object_type=object_type,
            type_mapping=type_mapping,
        )

    def _get_dataframe(
        self,
        filepath: str | Path,
        *args: Any
    ) -> pd.DataFrame:

        __df: pd.DataFrame = pd.read_csv(
            filepath
        )

        return __df.replace(np.nan, None)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..core import (
    core_data_object
)
from ..ena import (
    EnaDataSource,
    create_ena_datasource
)


def ena(**kwargs) -> EnaDataSource:
    ena = create_ena_datasource(
        ena_url=os.getenv('ENA_URL', Defaults.ENA_URL),
        ena_user=os.getenv('ENA_USER'),
        ena_password=os.getenv('ENA_PASSWORD'),
        ena_contact_name=os.getenv('ENA_CONTACT_NAME'),
        ena_contact_email=os.getenv('ENA_CONTACT_EMAIL'),
    )
    core_data_object(ena)
    return ena



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class StsManifestToElasticManifestConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        attributes = data_object.attributes.copy()
        attributes['project'] = data_object.project.id \
            if data_object.project is not None else None
        attributes['programme'] = data_object.project.programme \
            if data_object.project is not None else None
        attributes['status'] = data_object.manifest_status.status \
            if data_object.manifest_status is not None else None
        attributes['shipment_status'] = data_object.shipment_status.status \
            if data_object.shipment_status is not None else None
        attributes['compliance_status'] = data_object.compliance_status.status \
            if data_object.compliance_status is not None else None

        # wildlife_compliance_processors
        wcp_attribute = []
        if data_object.wildlife_compliance_processors is not None:
            for wcp in data_object.wildlife_compliance_processors:
                wcp_attribute.append(wcp.user.fullname)
        attributes['wildlife_compliance_processors'] = wcp_attribute

        ret = self._data_object_factory(
            'manifest',
            data_object.id,
            attributes=attributes,
            to_one={
                'sampleset': self._data_object_factory(
                    'sampleset',
                    data_object.sampleset.id
                ) if data_object.sampleset is not None else None,
            }
        )

        yield ret



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import urllib
from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import date
from typing import Any

from ..core import DataObject
from ..core.requested_fields import ReqFieldsTree

DocumentMeta = dict[str, Any]
DumpDict = dict[str, Any]
DumpDictMany = list[DumpDict]
ResponseDict = dict[str, DumpDict | DumpDictMany]


class View(ABC):
    """
    Provides an MVC-esque View class. Can serialize both an individual
    DataObject, as well as an Iterable of DataObject instances.
    """

    @abstractmethod
    def dump(
        self,
        data_object: DataObject,
        document_meta: DocumentMeta | None = None,
    ) -> ResponseDict:
        """
        Create a JSON:API response for an individual DataObject result
        """

    @abstractmethod
    def dump_bulk(
        self,
        data_objects: Iterable[DataObject],
        document_meta: DocumentMeta | None = None,
    ) -> ResponseDict:
        """
        Create a JSON:API response for an Iterable of DataObject results
        """


RelationshipDump = dict[str, dict[str, str]]
AllRelationshipsDump = dict[str, RelationshipDump]


class DefaultView(View):
    """
    Provides a default implementation of the View ABC.
    """

    def __init__(
        self,
        requested_tree: ReqFieldsTree,
        prefix: str = '',
        hop_limit: int | None = None,
    ) -> None:
        """
        Args:

        - prefix                - the URL prefix on which the
                                  data blueprint is served
        - hop_limit             - the maximum recursion limit
                                  on including related to-one
                                  objects. Default no limit
        - requested_tree        - a tree data structure of the
                                  requested fields for the query
        """

        self.__prefix = prefix
        self.__hop_limit = hop_limit
        self.__requested_tree = requested_tree

    def dump(
        self,
        data_object: DataObject,
        document_meta: DocumentMeta | None = None,
    ) -> ResponseDict:
        included = IncludedDumps()
        dumped = self.__dump_object(
            data_object,
            included,
            tree=self.__requested_tree,
        )
        response = {'data': dumped}
        if included:
            response['included'] = included.as_list()
        if document_meta is not None:
            response['meta'] = document_meta
        return response

    def dump_bulk(
        self,
        data_objects: Iterable[DataObject],
        document_meta: DocumentMeta | None = None,
    ) -> ResponseDict:
        included = IncludedDumps()
        dumped = [
            self.__dump_object(
                data_object,
                included,
                tree=self.__requested_tree,
            )
            for data_object in data_objects
        ]
        response = {'data': dumped}
        if included:
            response['included'] = included.as_list()
        if document_meta is not None:
            response['meta'] = document_meta
        return response

    def __dump_object(
        self,
        data_object: DataObject,
        included: IncludedDumps,
        tree: ReqFieldsTree,
    ) -> DumpDict:
        """
        Returns a JSON:API resource object for the `data_object`, recursively
        adding related objects as specified in the `tree: ReqFieldsTree`
        argument.  Related objects are accumulated in the `incldued` array.
        """
        dump = {'type': data_object.type, 'id': null_or_str(data_object.id)}
        # Stub trees are created by requested_fields paths ending in ".id"
        if not tree.is_stub:
            self.__add_attributes(data_object, dump, tree)
        if tree.has_relationships:
            self.__add_relationships(data_object, dump, included, tree)
        return dump

    def __add_attributes(
        self,
        data_object: DataObject,
        dump: DumpDict,
        tree: ReqFieldsTree | None,
    ):
        """
        If attributes are specified in the `tree: ReqFieldsTree`, adds only
        those to the dump.  Default is to add all attribtues.
        """
        if tree and (attr_names := tree.attribute_names):
            # Only add requested attributes
            dump['attributes'] = self.__convert_attributes(
                {name: getattr(data_object, name) for name in attr_names}
            )
        elif data_object.attributes:
            # Default behaviour is to add all attributes
            dump['attributes'] = self.__convert_attributes(data_object.attributes)

    def __add_relationships(
        self,
        data_object: DataObject,
        dump: DumpDict,
        included: IncludedDumps,
        tree: ReqFieldsTree | None = None,
    ) -> DumpDict:
        rel_dict = self.__dump_to_one_relationships(
            data_object, included, tree
        ) | self.__dump_to_many_relationships(data_object, included, tree)
        if rel_dict:
            dump['relationships'] = rel_dict

    def __dump_to_one_relationships(
        self,
        data_object: DataObject,
        included: IncludedDumps,
        tree: ReqFieldsTree,
    ) -> RelationshipDump:
        to_ones = {}
        for rel in tree.to_one_names():
            if rel in data_object._to_one_objects:
                one_dump = None
                if one := data_object._to_one_objects.get(rel):
                    one_dump = {'data': self.__dump_stub(one, rel)}
                    if sub_tree := tree.get_sub_tree(rel):
                        included.add_dump(self.__dump_object(one, included, tree=sub_tree))
                to_ones[rel] = one_dump
        return to_ones

    def __dump_to_many_relationships(
        self,
        data_object: DataObject,
        included: IncludedDumps,
        tree: ReqFieldsTree,
    ) -> RelationshipDump:
        oid = data_object.id
        quoted_id = None if oid is None else urllib.parse.quote(str(oid), safe='')
        to_many = {}
        for rel in tree.to_many_names():
            sub_tree = tree.get_sub_tree(rel)
            if sub_tree and rel in data_object._to_many_objects:
                many_obj = data_object._to_many_objects.get(rel)
                to_many[rel] = {'data': [self.__dump_stub(x, rel) for x in many_obj]}
                for obj in many_obj:
                    included.add_dump(self.__dump_object(obj, included, sub_tree))
            elif quoted_id:
                link = f'{self.__prefix}/{data_object.type}/{quoted_id}/{rel}'
                to_many[rel] = {'links': {'related': link}}
        return to_many

    def __dump_stub(self, obj: DataObject, rel_name: str) -> dict[str, str]:
        """
        Create a stub JSON:API object, known in the JSON:API spec as
        a "resource identifier object".  Contains a sanity check for the `id`
        attribute having a value.  If we want to support, for example,
        storing related objects with auto-incremented IDs, we will need to
        implement creating `lid` local IDs for linking to resource objects in
        the `included` array.
        """
        if obj.id is None:
            msg = (
                f"Cannot serialise '{obj.type}' object in relation"
                f" '{rel_name}' because it has no `id` attribute"
            )
            raise ValueError(msg)
        return {'type': obj.type, 'id': str(obj.id)}

    def __convert_attributes(self, attributes: dict[str, Any]) -> dict[str, Any]:
        return {k: self.__convert_value(v) for k, v in attributes.items()}

    def __convert_value(self, val: Any, /) -> Any:
        if isinstance(val, date):
            # `datetime` is a subclass of `date`
            return val.isoformat()
        return val


def null_or_str(oid: Any, /):
    """
    Return `oid` as a string if it isn't `None`
    """
    return None if oid is None else str(oid)


class IncludedDumps:
    """
    Maintains objects to be returned in the JSON:API `included` list, indexed
    by tuples of `(type, id)`.
    """

    def __init__(self):
        self.__type_id: dict[tuple[str, str], DumpDict] = {}

    def __len__(self):
        """
        Implemented so that an `IncludedDumps` object returns true in boolean
        context when it has entries.
        """
        return len(self.__type_id)

    def as_list(self):
        return list(self.__type_id.values())

    def add_dump(self, dump: DumpDict):
        """
        Add a new DumpDict to the collection.
        """
        key = dump['type'], dump['id']
        if key not in self.__type_id:
            self.__type_id[key] = dump



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

__AndFilterTerm = dict[
    str,
    dict[str, Any]
]


ExactFilter = Dict[str, Any]
ContainsFilter = Dict[str, str]
InListFilter = Dict[str, List[Any]]
RangeFilter = Dict[str, Dict[str, Any]]
AndFilter = dict[
    str,
    __AndFilterTerm
]


@dataclass
class DataSourceFilter:
    """
    Configures the filtering for a DataSource
    """
    exact: Optional[ExactFilter] = None
    contains: Optional[ContainsFilter] = None
    in_list: Optional[InListFilter] = None
    range: Optional[RangeFilter] = None  # noqa A003
    and_: Optional[AndFilter] = None



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

def _heading(heading_number, title):
    hashes = '#' * heading_number
    return f'{hashes} {title}'


def h1(title):
    return _heading(1, title)


def h2(title):
    return _heading(2, title)


def h3(title):
    return _heading(3, title)


def h4(title):
    return _heading(4, title)


def h5(title):
    return _heading(5, title)


def h6(title):
    return _heading(6, title)


def h7(title):
    return _heading(7, title)


def link(url, text):
    return f'[{text}]({url})'



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class StsSamplesetToElasticSamplesetConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        attributes = data_object.attributes
        attributes['gal_abbreviation'] = data_object.gal.abbreviation \
            if data_object.gal is not None else None
        attributes['project'] = data_object.project.id \
            if data_object.project is not None else None
        attributes['programme'] = data_object.project.programme \
            if data_object.project is not None else None
        attributes['status'] = data_object.sampleset_status.status \
            if data_object.sampleset_status is not None else None

        # SamplesetLegal is a one-to-many but we treat it as a one-to-one
        if data_object.sampleset_legals is not None:
            try:
                legal = next(data_object.sampleset_legals)
                attributes['legal_contract'] = legal.contract
                attributes['legal_reference'] = legal.reference
                attributes['legal_status'] = legal.status
                attributes['legal_comment'] = legal.comment
                attributes['legal_status_updated_at'] = \
                    legal.status_updated_at
            except StopIteration:
                pass

        # legal_compliance_processors
        lcp_attribute = []
        if data_object.legal_compliance_processors is not None:
            for lcp in data_object.legal_compliance_processors:
                lcp_attribute.append(lcp.user.fullname)
        attributes['legal_compliance_processors'] = lcp_attribute

        # sampleset_research_governance_processors
        rgp_attribute = []
        if data_object.sampleset_research_governance_processors is not None:
            for rgp in data_object.sampleset_research_governance_processors:
                rgp_attribute.append(rgp.user.fullname)
        attributes['research_governance_processors'] = rgp_attribute

        # sampleset_managers
        ssm_attribute = []
        if data_object.sampleset_managers is not None:
            for ssm in data_object.sampleset_managers:
                ssm_attribute.append(ssm.user.fullname)
        attributes['managers'] = ssm_attribute

        # There may be multiple SamplesetResearchGovernance each with different type
        if data_object.sampleset_research_governances is not None:
            for rg in data_object.sampleset_research_governances:
                research_governance_type = rg.research_governance_type.lower()
                attributes[f'rg_status_{research_governance_type}'] = \
                    rg.compliance_status.status if rg.compliance_status is not None else None
                attributes[f'rg_status_updated_at_{research_governance_type}'] = \
                    rg.updated_at
        ret = self._data_object_factory(
            'sampleset',
            data_object.id,
            attributes=attributes
        )
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from typing import Any, Callable

if typing.TYPE_CHECKING:
    from .model import Model


def ext(
    cls: type[Model] | None = None,
    *,
    target: str = 'ext'
) -> Callable[[type[Model]], type[Model]] | type[Model]:
    """
    This column supports a (for now ReadOnly) dynamic "promotion"
    of entries, from a `target` column that produces a `dict` value,
    to top-level entries in the property `Model().instance_attributes`.

    Can be decorated either with or without parentheses, the latter of
    which supports overriding the name of the `target` column.
    """

    if cls is not None:

        def exclude_wrapper(
            cls_fn: Callable[[], list[str]]
        ) -> Callable[[], list[str]]:

            def inner() -> list[str]:
                excluded = cls_fn()
                return [
                    *excluded,
                    target
                ]

            return inner

        cls.get_excluded_column_names = exclude_wrapper(
            cls.get_excluded_column_names
        )

        def attrs_wrapper(_property: property) -> property:

            @property
            def inner(self: Model) -> dict[str, Any]:
                ext_attrs = getattr(self, target)
                return {
                    **(ext_attrs if ext_attrs else {}),
                    **_property.fget(self)
                }

            return inner

        cls.instance_attributes = attrs_wrapper(
            cls.instance_attributes
        )

        return cls

    def decorator(cls: type[Model]) -> type[Model]:
        return ext(cls, target=target)

    return decorator



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import importlib
from dataclasses import dataclass

from tol.core import DataObject
from tol.core.factory import DataObjectFactory
from tol.core.validate import ValidationResult, Validator


class ConverterAndValidateValidator(Validator):
    """
    Convert DataObjects, validate the converted ones, and return the original
    input unchanged. Inner validator results are not merged here.

    {
        "converters": [{
            "module": "<path.to.module>",
            "class_name": "<path.to.ConverterClass>",
            "config_details": { ... }
        }],
        "validators": [{
            "module": "<path.to.module>",
            "class_name": "<path.to.ValidatorClass>",
            "config_details": { ... }
        }]
    }

    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        converters: list[dict]
        validators: list[dict]

    __slots__ = [
        '__converters',
        '__validators'
    ]

    def __init__(
        self,
        config: Config,
        data_object_factory: DataObjectFactory,
        **kwargs
    ) -> None:
        super().__init__()
        self.__converters = []
        self.__validators = []

        for conv in config.converters:
            __module = importlib.import_module(conv.get('module'))
            converter_class = getattr(__module, conv.get('class_name'))

            converter_conf = converter_class.Config(
                **conv.get('config_details')
            )
            self.__converters.append(converter_class(
                data_object_factory=data_object_factory,
                config=converter_conf,
            ))
        for val in config.validators:
            __module = importlib.import_module(val.get('module'))
            validator_class = getattr(__module, val.get('class_name'))

            validator_conf = validator_class.Config(
                **val.get('config_details')
            )
            self.__validators.append(validator_class(
                data_object_factory=data_object_factory,
                config=validator_conf,
            ))

    def _validate_data_object(self, obj: DataObject) -> None:
        converted_objs = [obj]
        for converter in self.__converters:
            converted_objs = converter.convert_iterable(converted_objs)
        for obj in converted_objs:
            for validator in self.__validators:
                validator._validate_data_object(obj)

    @property
    def results(self) -> list[ValidationResult]:
        return [result for validator in self.__validators for result in validator.results]

    @property
    def warnings(self) -> list[ValidationResult]:
        return [warning for validator in self.__validators for warning in validator.warnings]

    @property
    def errors(self) -> list[ValidationResult]:
        return [error for validator in self.__validators for error in validator.errors]



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cached_property, reduce, wraps
from typing import Any, Callable

from .datasource_error import DataSourceError
from .operator import (
    ALL_OPERATORS,
    get_operator_member_names,
)

if typing.TYPE_CHECKING:
    from .datasource import OperableDataSource


class DataSourceSession:
    """
    A session for a `DataSource` instance.

    Supports all operations that its `__host` does.
    """

    def __init__(
        self,
        data_source: OperableDataSource
    ) -> None:

        self.__host = data_source

    def __getattribute__(self, __name: str) -> Any:
        if __name.startswith('_'):
            return object.__getattribute__(self, __name)
        else:
            return self.__proxy(__name)

    def __enter__(self) -> OperableSession:
        return self

    def __exit__(self, type_, value, tb) -> None:
        pass

    def __proxy(
        self,
        __name: str
    ) -> Any:
        """Proxies non-local names to `.__host`"""

        self.__validate_proxy(__name)

        proxied = getattr(
            self.__host,
            __name
        )

        if __name not in self.__all_operator_method_dict:
            return proxied

        if not callable(proxied):
            return proxied

        return self.__decorate_add_kwarg(proxied)

    def __decorate_add_kwarg(
        self,
        operator_method: Callable
    ) -> Callable:
        """
        Inserts this instance as the `session` kwarg to the
        given `Callable`'s invocation.
        """

        @wraps(operator_method)
        def wrapper(*args, **kwargs) -> Any:
            return operator_method(
                *args,
                **kwargs,
                session=self
            )

        return wrapper

    def __validate_proxy(self, __name: str) -> None:
        """
        Confirms that the member is implemented
        if it belongs to an `Operator`.
        """

        if __name in self.__all_operator_method_dict:
            operator_name = self.__all_operator_method_dict[__name]

            if operator_name not in self.__operator_names:
                raise DataSourceError(
                    title='Unsupported Operation',
                    detail=(
                        'This Operator method is not implemented '
                        'by the hosting DataSource.'
                    ),
                    status_code=500,
                )

    @cached_property
    def __operator_names(self) -> list[str]:
        """
        A `list` of classnames of implemented operators
        """

        return [
            o.__name__ for o in ALL_OPERATORS
            if isinstance(
                self.__host,
                o
            )
        ]

    @cached_property
    def __all_operator_method_dict(
        self
    ) -> dict[str, str]:
        """
        Maps operation method names to their
        `Operator` name (all, not just
        those implemented)
        """

        inverted = (
            (
                o.__name__,
                get_operator_member_names(o),
            )
            for o in ALL_OPERATORS
        )

        def __act(
            d: dict[str, str],
            o: str,
            names: list[str]
        ) -> dict[str, str]:

            add_dict = {
                n: o for n in names
            }
            return d | add_dict

        return reduce(
            lambda d, p: __act(d, *p),
            inverted,
            {}
        )


if typing.TYPE_CHECKING:

    class OperableSession(
        OperableDataSource,
        DataSourceSession,
    ):
        """A type hint. For inheriting, use `DataSourceSession`."""



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing

from ..core import DataSourceError

if typing.TYPE_CHECKING:
    from .model import Model


class BadColumnError(DataSourceError):
    def __init__(self, model: Model, column_name: str):
        detail = (
            f'The column with name "{column_name}" does not exist '
            f'on the model with tablename "{model.get_table_name()}".'
        )
        super().__init__('Bad Column', detail, 400)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


from .blueprint import DbAuthBlueprint, db_auth_blueprint  # noqa F401
from .models import ModelTuple  # noqa F401
